//********************************************************************************************************************************
//*                                                                                                                              *
//*     File Operations Library 1.0.10.10  3delite 2023-2024                                                                    *
//*     See File Operations Library ReadMe.txt for details                                                                       *
//*                                                                                                                              *
//* Licenses available if you like and use this library:                                                                         *
//* Freeware License: 15                                                                                                        *
//*     http://www.shareit.com/product.html?productid=301063451                                                                  *
//* Shareware License: 25                                                                                                       *
//*     http://www.shareit.com/product.html?productid=301063452                                                                  *
//* Commercial License: 95                                                                                                      *
//*     http://www.shareit.com/product.html?productid=301063450                                                                  *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/FileOperationsLibrary.html                                *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************

unit FileOperationsCalculateFolderSize;

{$WARN SYMBOL_PLATFORM OFF}

interface

Uses
    SysUtils,
    Classes,
    Windows,
    SHLObj;

Const
    STR_MSG_FOLDER_SIZE_RESULT = 'Calculate folder size result';

type
    TCalculateFolderSizeItem = class
    public
        FolderName: String;
        PIDL: PItemIDList;
        Size: UInt64;
        ListingID: Int64;
        Empty: Boolean;
        Destructor Destroy; override;
    end;
    TCalculateFolderSizeItems = Array of TCalculateFolderSizeItem;

type
    TCalculateFolderSizeResult = record
        Size: UInt64;
        Empty: Boolean;
    end;

type
    TCalculateFolderSizeThread = class(TThread)
    private
        { Private declarations }
        FFolderSizeItems: TCalculateFolderSizeItems;
        FHandle: HWND;
    protected
        procedure Execute; override;
    public
        Constructor Create(FolderSizeItems: TCalculateFolderSizeItems; Handle: HWND);
        Destructor Destroy; override;
        function CalculateFolderSizeFile(FolderName: String): TCalculateFolderSizeResult;
        function CalculateFolderSizePIDL(PIDL: PItemIDList): TCalculateFolderSizeResult;
    end;

var
    MSG_FOLDER_SIZE_RESULT: Cardinal;

implementation

Uses
    ActiveX,
    Winapi.PropKey,
    FileOperationsLibrary;

function GetFileNameFromPIDL(Item: PItemIDList): String;
var
    FBuf: array[0..MAX_PATH] of Char;
begin
    SHGetPathFromIDList(Item, @FBuf[0]);
    Result := FBuf;
end;

function GetPIDLFileSize(Item: PItemIDList): UInt64;
var
    Error: HRESULT;
    SiItem: IShellItem;
    SiItem2: IShellItem2;
begin
    Result := 0;
    Error := SHCreateItemFromIDList(Item, IShellItem, SiItem);
    if Succeeded(Error) then begin
        SiItem2 := SiItem as IShellItem2;
        SiItem2.GetUInt64(PKEY_Size, Result);
    end;
end;

{ TCalculateFolderSizeThread }

function TCalculateFolderSizeThread.CalculateFolderSizeFile(FolderName: String): TCalculateFolderSizeResult;
var
    SearchRec: TSearchRec;
begin
    Result.Size := 0;
    Result.Empty := True;
    if Terminated then begin
        Exit;
    end;
    try
        if FindFirst(IncludeTrailingPathDelimiter(FolderName) + '*.*', faAnyFile, SearchRec) = 0 then begin
            repeat
                if (SearchRec.Name <> '.')
                AND (SearchRec.Name <> '..')
                then begin
                    if (SearchRec.FindData.dwFileAttributes AND FILE_ATTRIBUTE_DIRECTORY) = 0 then begin
                        if SearchRec.Size > 0 then begin
                            Result.Empty := False;
                            Result.Size := Result.Size + UInt64(SearchRec.Size);
                        end;
                    end else begin
                        Result.Empty := False;
                        Result.Size := Result.Size + CalculateFolderSizeFile(IncludeTrailingPathDelimiter(FolderName) + SearchRec.Name).Size;
                    end;
                end;
                if Terminated then begin
                    Exit;
                end;
            until FindNext(SearchRec) <> 0;
        end;
    finally
        System.SysUtils.FindClose(SearchRec);
    end;
end;

function TCalculateFolderSizeThread.CalculateFolderSizePIDL(PIDL: PItemIDList): TCalculateFolderSizeResult;
var
    Desktop: IShellFolder;
    EnumIDLIst: IEnumIDList;
    Fetched: ULONG;
    CurPIDL: PItemIDList;
    NewPIDL: PItemIDList;
    Attributes: Cardinal;
    Error: Integer;
    Flags: Cardinal;
    CurrentShellFolder: IShellFolder;
    FolderName: String;
    ItemName: String;
    PIDLFileSize: UInt64;
    FileSize: Int64;
begin
    Result.Size := 0;
    Result.Empty := True;
    if Terminated then begin
        Exit;
    end;
    //* Item types
    if SHGetDesktopFolder(Desktop) = NOERROR then begin
        Flags := SHCONTF_FOLDERS OR SHCONTF_NONFOLDERS OR SHCONTF_INCLUDEHIDDEN;
        Error := Desktop.BindToObject(PIDL, nil, IID_IShellFolder, Pointer(CurrentShellFolder));
        if Error = NOERROR then begin
            CurrentShellFolder.EnumObjects(Handle, Flags, EnumIDList);
            if Assigned(EnumIDList) then begin
                while EnumIDList.Next(1, CurPIDL, Fetched) = S_OK do begin
                    try
                        Attributes := SFGAO_FOLDER OR SFGAO_FILESYSANCESTOR OR SFGAO_FILESYSTEM {OR SFGAO_COMPRESSED};
                        CurrentShellFolder.GetAttributesOf(1, CurPIDL, Attributes);
                        NewPIDL := ILCombine(PIDL, CurPIDL);
                        try
                            if ((Attributes AND SFGAO_FOLDER) > 0) then begin
                                FolderName := GetFileNameFromPIDL(NewPIDL);
                                if DirectoryExists(FolderName) then begin
                                    Result.Empty := False;
                                    Result.Size := Result.Size + CalculateFolderSizeFile(FolderName).Size;
                                end else begin
                                    if FileExists(FolderName) then begin
                                        PIDLFileSize := NGetFileSize(ItemName);
                                    end else begin
                                        PIDLFileSize := GetPIDLFileSize(NewPIDL);
                                    end;
                                    //* Not archive if file size is 0
                                    if PIDLFileSize = 0 then begin
                                        Result.Empty := False;
                                        Result.Size := Result.Size + CalculateFolderSizePIDL(NewPIDL).Size;
                                    end else begin
                                        Result.Empty := False;
                                        Result.Size := Result.Size + PIDLFileSize;
                                    end;
                                end;
                            end else begin
                                ItemName := GetFileNameFromPIDL(NewPIDL);
                                if FileExists(ItemName) then begin
                                    FileSize := NGetFileSize(ItemName);
                                    if FileSize > 0 then begin
                                        Result.Empty := False;
                                        Result.Size := Result.Size + UInt64(FileSize);
                                    end;
                                end else begin
                                    Result.Empty := False;
                                    Result.Size := Result.Size + GetPIDLFileSize(NewPIDL);
                                end;
                            end;
                        finally
                            CoTaskMemFree(NewPIDL);
                        end;
                    finally
                        CoTaskMemFree(CurPIDL);
                    end;
                    if Terminated then begin
                        Exit;
                    end;
                end;
            end;
        end;
    end;
end;

constructor TCalculateFolderSizeThread.Create(FolderSizeItems: TCalculateFolderSizeItems; Handle: HWND);
begin
    inherited Create(False);
    FreeOnTerminate := False;
    FFolderSizeItems := FolderSizeItems;
    FHandle := Handle;
end;

destructor TCalculateFolderSizeThread.Destroy;
var
    i: Integer;
begin
    for i := Low(FFolderSizeItems) to High(FFolderSizeItems) do begin
        FreeAndNil(FFolderSizeItems[i]);
    end;
    inherited;
end;

procedure TCalculateFolderSizeThread.Execute;
var
    CalculateFolderSizeResult: TCalculateFolderSizeResult;
begin
    inherited;
    //i := 0;
    while Length(FFolderSizeItems) > 0 do begin
        //FolderSize := 0;
        if DirectoryExists(FFolderSizeItems[0].FolderName) then begin
            CalculateFolderSizeResult := CalculateFolderSizeFile(FFolderSizeItems[0].FolderName);
        end else begin
            CalculateFolderSizeResult := CalculateFolderSizePIDL(FFolderSizeItems[0].PIDL);
        end;
        if Terminated then begin
            Break;
        end;
        FFolderSizeItems[0].Size := CalculateFolderSizeResult.Size;
        FFolderSizeItems[0].Empty := CalculateFolderSizeResult.Empty;
        PostMessage(FHandle, MSG_FOLDER_SIZE_RESULT, WPARAM(FFolderSizeItems[0]), 0);
        Delete(FFolderSizeItems, 0, 1);
        if NOT Assigned(FFolderSizeItems) then begin
            Break;
        end;
    end;
end;

{ TCalculateFolderSizeItem }

destructor TCalculateFolderSizeItem.Destroy;
begin
    if Assigned(PIDL) then begin
        CoTaskMemFree(PIDL);
        PIDL := nil;
    end;
    inherited;
end;

Initialization

    MSG_FOLDER_SIZE_RESULT := RegisterWindowMessage(STR_MSG_FOLDER_SIZE_RESULT);

end.
