//********************************************************************************************************************************
//*                                                                                                                              *
//*     File Operations Library 1.0.10.10  3delite 2023-2024                                                                    *
//*     See File Operations Library ReadMe.txt for details                                                                       *
//*                                                                                                                              *
//* Licenses available if you like and use this library:                                                                         *
//* Freeware License: 15                                                                                                        *
//*     http://www.shareit.com/product.html?productid=301063451                                                                  *
//* Shareware License: 25                                                                                                       *
//*     http://www.shareit.com/product.html?productid=301063452                                                                  *
//* Commercial License: 95                                                                                                      *
//*     http://www.shareit.com/product.html?productid=301063450                                                                  *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/FileOperationsLibrary.html                                *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************

unit FileOperationsLibrary;

{$WARN SYMBOL_PLATFORM OFF}

interface

Uses
    SysUtils,
    Classes,
    Types,
    Windows,
    SHLObj,
    System.Zip;

Const
    //MOVEEFFECTBYTES: packed Array [0..3] of Byte = ($02, $00, $00, $00);
    MOVEEFFECTBYTESVALUE: Cardinal = $00000002;

type
    TFOFileList = TStringDynArray;
    //PTArrayOfPItemIDList = ^TArrayOfPItemIDList;
    TArrayOfPItemIDList = array of PItemIDList;

type
    TZIPProgressCompleted = procedure(FileName: String; FileSize: UInt64) of object;
    TZIPOperationTotalBytes = procedure(TotalBytesSize: UInt64) of object;

type
    TFOCopyFilesThread = class(TThread)
    private
        FDestinationFolder: String;
        FFiles: TFOFileList;
        FDestinationFolderPIDL: PItemIDList;
        FItemIdList: TArrayOfPItemIDList;
        FHandle: THandle;
    protected
        procedure Execute; override;
    public
        Constructor Create(DestinationFolder: String; Files: TFOFileList; Handle: THandle); overload;
        Constructor Create(DestinationFolder: PItemIDList; ItemIdList: TArrayOfPItemIDList; Handle: THandle); overload;
        Destructor Destroy; override;
    end;

type
    TFOMoveFilesThread = class(TThread)
    private
        FDestinationFolder: String;
        FFiles: TFOFileList;
        FDestinationFolderPIDL: PItemIDList;
        FItemIdList: TArrayOfPItemIDList;
        FHandle: THandle;
    protected
        procedure Execute; override;
    public
        Constructor Create(DestinationFolder: String; Files: TFOFileList; Handle: THandle); overload;
        Constructor Create(DestinationFolder: PItemIDList; ItemIdList: TArrayOfPItemIDList; Handle: THandle); overload;
        Destructor Destroy; override;
    end;

type
    TFODeleteFilesThread = class(TThread)
    private
        FFiles: TFOFileList;
        FPermanently: Boolean;
        FItemIdList: TArrayOfPItemIDList;
        FHandle: THandle;
    protected
        procedure Execute; override;
    public
        Constructor Create(Files: TFOFileList; Permanently: Boolean; Handle: THandle); overload;
        Constructor Create(ItemIdList: TArrayOfPItemIDList; Permanently: Boolean; Handle: THandle); overload;
        Destructor Destroy; override;
    end;

    function GetFileNameFromPIDL(Item: PItemIDList): String;
    function GetPIDLFromFileName(FileName: String): PItemIDList;
    function GetPIDLDisplayName(Item: PItemIDList): String;
    function GetPIDLAttributes(Item: PItemIDList; var Attributes: Cardinal): HResult;

    function NGetFileSize(FileName: String): UInt64;

    procedure CopyPIDLList(SourcePItemIDList: TArrayOfPItemIDList; var DestinationPItemIDList: TArrayOfPItemIDList);
    procedure FreePIDLList(var ArrayOfPItemIDList: TArrayOfPItemIDList);

    function FOCopyFile(DestinationFolder: String; FileName: String; Handle: THandle): HRESULT; overload;
    function FOCopyFile(DestinationFolder: PItemIDList; Item: PItemIDList; Handle: THandle): HRESULT; overload;

    function FOMoveFile(DestinationFolder: String; FileName: String; Handle: THandle): HRESULT; overload;
    function FOMoveFile(DestinationFolder: PItemIDList; Item: PItemIDList; Handle: THandle): HRESULT; overload;

    function FODeleteFile(FileItem: String; Permanently: Boolean; Handle: THandle): HRESULT; overload;
    function FODeleteFile(Item: PItemIDList; Permanently: Boolean; Handle: THandle): HRESULT; overload;

    function FOCopyFiles(DestinationFolder: String; Files: TFOFileList; Handle: THandle): HRESULT; overload;
    function FOCopyFiles(DestinationFolder: PItemIDList; ItemIdList: TArrayOfPItemIDList; Handle: THandle): HRESULT; overload;

    function FOMoveFiles(DestinationFolder: String; Files: TFOFileList; Handle: THandle): HRESULT; overload;
    function FOMoveFiles(DestinationFolder: PItemIDList; ItemIdList: TArrayOfPItemIDList; Handle: THandle): HRESULT; overload;

    function FODeleteFiles(Files: TFOFileList; Permanently: Boolean; Handle: THandle): HRESULT; overload;
    function FODeleteFiles(ItemIdList: TArrayOfPItemIDList; Permanently: Boolean; Handle: THandle): HRESULT; overload;

    function FOCopyFileAs(DestinationFolder: String; FileName: String; NewFileName: String; Handle: THandle): HRESULT; overload;
    function FOCopyFileAs(DestinationFolder: PItemIDList; FileName: PItemIDList; NewFileName: String; Handle: THandle): HRESULT; overload;

    function FOMoveFileAs(DestinationFolder: String; FileName: String; NewFileName: String; Handle: THandle): HRESULT; overload;
    function FOMoveFileAs(DestinationFolder: PItemIDList; FileName: PItemIDList; NewFileName: String; Handle: THandle): HRESULT; overload;

    function FOCreateFolder(DestinationFolder: PItemIDList; FolderName: String; Handle: THandle; var IsZIP: Boolean): HRESULT;

    function IsZIPFolder(PIDL: PItemIDList): Boolean; overload;
    function IsZIPFolder(PIDL: PItemIDList; var ZIPFileName: String): Boolean; overload;
    function ZIPFolderPath(PIDL: PItemIDList): String;

    function FOCopyFilesToZIP(DestinationFolder: PItemIDList; ItemIdList: TArrayOfPItemIDList; Handle: THandle; ZipProgressEvent: TZipProgressEvent = nil; ZIPProgressCompleted: TZIPProgressCompleted = nil): HRESULT; overload;
    function FOMoveFilesToZIP(DestinationFolder: PItemIDList; ItemIdList: TArrayOfPItemIDList; Handle: THandle; ZipProgressEvent: TZipProgressEvent = nil; ZIPProgressCompleted: TZIPProgressCompleted = nil): HRESULT; overload;
    function FOCopyFileToZIPAs(DestinationFolder: PItemIDList; Item: PItemIDList; NewFileName: String; Handle: THandle; ZipProgressEvent: TZipProgressEvent = nil; ZIPProgressCompleted: TZIPProgressCompleted = nil): HRESULT; overload;
    function FOMoveFileToZIPAs(DestinationFolder: PItemIDList; Item: PItemIDList; NewFileName: String; Handle: THandle; ZipProgressEvent: TZipProgressEvent = nil; ZIPProgressCompleted: TZIPProgressCompleted = nil): HRESULT; overload;
    function FODeleteFilesInZIP(ItemIdList: TArrayOfPItemIDList; Handle: THandle; ZipProgressEvent: TZipProgressEvent = nil; ZIPProgressCompleted: TZIPProgressCompleted = nil): HRESULT; overload;
    function FODeleteFileInZIP(Item: PItemIDList; Handle: THandle; ZipProgressEvent: TZipProgressEvent = nil; ZIPProgressCompleted: TZIPProgressCompleted = nil): HRESULT; overload;

    function CalculateFolderSizeFileSystem(FolderName: String): UInt64;
    function CalculateFolderSizePIDL(PIDL: PItemIDList; Handle: HWND): UInt64;
    function CalculateFolderSizeFileSystemList(FileList: TFOFileList; Handle: HWND): UInt64;
    function CalculateFolderSizeItemIDList(ItemIdList: TArrayOfPItemIDList; Handle: HWND): UInt64;

    procedure CopyFileNamesToClipboard(FileNames: TFOFileList; CutMode: Boolean);
    procedure CopyPIDLsToClipboard(ItemIdList: TArrayOfPItemIDList; CutMode: Boolean);
    function GetPIDLsFromClipboard(var ItemIdList: TArrayOfPItemIDList; var CutMode: Boolean): Boolean;

implementation

Uses
    ComObj,
    ShellAPI,
    ActiveX,
    Winapi.PropKey,
    IOUtils,
    KnownFolders,
    ClipBrd,
    Dialogs,
    System.UITypes,
    Vcl.Controls;

var
    CF_SHELLIDLIST: UINT = 0;
    CF_PREFERREDDROPEFFECT: UINT = 0;

type
    TZIPFileHelper = class helper for TZipFile
    public
        function CheckToCreatePath(ArchiveFileName: String): Boolean;
        procedure Add(const FileName: string; const ArchiveFileName: string = ''; Compression: TZipCompression = zcDeflate); overload;
        //procedure Add(Data: TBytes; const ArchiveFileName: string; Compression: TZipCompression = zcDeflate); overload;
    end;

function TZIPFileHelper.CheckToCreatePath(ArchiveFileName: String): Boolean;
var
    FolderPath: String;
    EndPosition: Integer;
    Found: Boolean;
    i: Integer;
    Data: TBytes;
begin
    Result := False;
    EndPosition := 0;
    repeat
        Found := False;
        EndPosition := Pos('/', ArchiveFileName, EndPosition + 1);
        if EndPosition > 0 then begin
            FolderPath := Copy(ArchiveFileName, 1, EndPosition);
            for i := 0 to FileCount - 1 do begin
                if SameText(FileName[i], FolderPath) then begin
                    Found := True;
                    Break;
                end;
            end;
            if NOT Found then begin
                inherited Add(Data, FolderPath);
                Result := True;
            end;
        end;
    until EndPosition = 0;
end;

procedure TZIPFileHelper.Add(const FileName: string; const ArchiveFileName: string = ''; Compression: TZipCompression = zcDeflate);
begin
    CheckToCreatePath(ArchiveFileName);
    inherited Add(FileName, ArchiveFileName, Compression);
end;

{
procedure TZIPFileHelper.Add(Data: TBytes; const ArchiveFileName: string; Compression: TZipCompression = zcDeflate);
begin
    CheckToCreatePath(ArchiveFileName);
    //inherited Add(Data, ArchiveFileName, Compression);
end;
}

function GetFileNameFromPIDL(Item: PItemIDList): String;
var
    FBuf: array[0..MAX_PATH] of Char;
begin
    Result := '';
    if NOT Assigned(Item) then begin
        Exit;
    end;
    SHGetPathFromIDList(Item, @FBuf[0]);
    Result := FBuf;
end;

function GetPIDLFromFileName(FileName: String): PItemIDList;
var
    Desktop: IShellFolder;
    OleStr: array[0..MAX_PATH] of WideChar;
    CardOut: Cardinal;
begin
     Result := nil;
     if Failed(SHGetDesktopFolder(Desktop)) then begin
        Exit;
     end;
     CopyMemory(@OleStr[0], @FileName[1], (Length(FileName) + 1) * 2);
     Desktop.ParseDisplayName(0, nil, @OleStr, CardOut, Result, CardOut);
end;

function GetPIDLDisplayName(Item: PItemIDList): String;
var
    Str: TStrRet;
    piFolder: IShellFolder;
begin
    Result := '';
    if Failed(SHGetDesktopFolder(piFolder)) then begin
        Exit;
    end;
    if piFolder.GetDisplayNameOf(Item, 0, Str) = NOERROR then begin
        Result := str.pOleStr;
        CoTaskMemFree(str.pOleStr);
    end;
end;

function GetPIDLAttributes(Item: PItemIDList; var Attributes: Cardinal): HResult;
var
    ParentFolder: IShellFolder;
    RelativePIDL: PItemIDList;
begin
    Result := SHBindToParent(Item, IID_IShellFolder, Pointer(ParentFolder), RelativePIDL);
    if Result = NOERROR then begin
        Attributes := SFGAO_FOLDER OR SFGAO_FILESYSANCESTOR OR SFGAO_FILESYSTEM OR SFGAO_CANRENAME OR SFGAO_CANDELETE {OR SFGAO_COMPRESSED};
        ParentFolder.GetAttributesOf(1, RelativePIDL, Attributes);
    end;
end;

procedure CopyPIDLList(SourcePItemIDList: TArrayOfPItemIDList; var DestinationPItemIDList: TArrayOfPItemIDList);
var
    i: Integer;
    PPIDL: PByte;
begin
    if Assigned(SourcePItemIDList) then begin
        SetLength(DestinationPItemIDList, Length(SourcePItemIDList));
        PPIDL := PByte(SourcePItemIDList);
        for i := 0 to Length(SourcePItemIDList) - 1 do begin
            DestinationPItemIDList[i] := ILClone(PItemIDList(PPIDL));
            Inc(PPIDL, ILGetSize(PItemIDList(PPIDL)));
        end;
    end;
end;

procedure FreePIDLList(var ArrayOfPItemIDList: TArrayOfPItemIDList);
var
    i: Integer;
begin
    if Assigned(ArrayOfPItemIDList) then begin
        for i := 0 to Length(ArrayOfPItemIDList) - 1 do begin
            CoTaskMemFree(ArrayOfPItemIDList[i]);
        end;
        SetLength(ArrayOfPItemIDList, 0);
    end;
end;

function NGetFileSize(FileName: String): UInt64;
var
  f: TSearchRec;
begin
    Result := 0;
    if FileExists(FileName) then begin
    	try
            if FindFirst(FileName, faAnyFile, f) = 0 then begin
                if f.Size > 0 then begin
                    Result := UInt64(f.Size);
                end;
            end;
        finally
            SysUtils.FindClose(f);
        end;
    end;
end;

function ExtractFilePathZIP(const FileName: string): string;
var
  I: Integer;
begin
  I := FileName.LastDelimiter(['/', DriveDelim]);
  Result := Copy(FileName, 1, I + 1);
end;

{ TFOCopyFilesThread }

function FOCopyFiles(DestinationFolder: String; Files: TFOFileList; Handle: THandle): HRESULT;
var
    i: Integer;
    FileOp: IFileOperation;
    SiDestFolder: IShellItem;
    ppsiItemArray: IShellItemArray;
    ItemIdList: TArrayOfPItemIDList;
    FileCount: Integer;
begin
    Result := S_FALSE;
    if Length(Files) = 0 then begin
        Exit;
    end;
    //create IFileOperation interface
    Result := CoCreateInstance(CLSID_FileOperation, nil, CLSCTX_ALL, IFileOperation, FileOp);
    if Succeeded(Result) then begin
        //set operations flags
        Result := FileOp.SetOperationFlags({FOF_NOCONFIRMATION OR} FOFX_NOMINIMIZEBOX);
        if Succeeded(Result) then begin
            //get destination folder shell item
            Result := SHCreateItemFromParsingName(PChar(DestinationFolder), nil, IShellItem, SiDestFolder);
            if Succeeded(Result) then begin
                FileCount := Length(Files);
                //* Our PIDL array
                SetLength(ItemIdList, Length(Files));
                for i := 0 to Length(Files) - 1 do begin
                    ItemIdList[i] := GetPIDLFromFileName(Files[i]);
                end;
                //* Create an array interface
                Result := SHCreateShellItemArrayFromIDLists(FileCount, PCUIDLIST_ABSOLUTE_ARRAY(ItemIdList), ppsiItemArray);
                if Succeeded(Result) then begin
                    //* Assign copy items task
                    Result := FileOp.CopyItems(ppsiItemArray, SiDestFolder);
                    if Succeeded(Result) then begin
                        //* Do the copy task
                        Result := FileOp.PerformOperations;
                    end;
                end;
                //* Free all source file PIDLs
                for i := 0 to Length(Files) - 1 do begin
                    CoTaskMemFree(ItemIdList[i]);
                end;
            end;
        end;
        //OleCheck(Result);
    end;
end;

function FOCopyFiles(DestinationFolder: PItemIDList; ItemIdList: TArrayOfPItemIDList; Handle: THandle): HRESULT;
var
    FileOp: IFileOperation;
    SiDestFolder: IShellItem;
    ppsiItemArray: IShellItemArray;
begin
    Result := S_FALSE;
    if (Length(ItemIdList) = 0)
    OR (NOT Assigned(DestinationFolder))
    OR (NOT Assigned(ItemIdList))
    then begin
        Exit;
    end;
    //create IFileOperation interface
    Result := CoCreateInstance(CLSID_FileOperation, nil, CLSCTX_ALL, IFileOperation, FileOp);
    if Succeeded(Result) then begin
        //set operations flags
        Result := FileOp.SetOperationFlags({FOF_NOCONFIRMATION OR} FOFX_NOMINIMIZEBOX);
        if Succeeded(Result) then begin
            //get destination folder shell item
            Result := SHCreateItemFromIDList(DestinationFolder, IShellItem, SiDestFolder);
            if Succeeded(Result) then begin
                //* Create an array interface
                Result := SHCreateShellItemArrayFromIDLists(Length(ItemIdList), PCUIDLIST_ABSOLUTE_ARRAY(ItemIdList), ppsiItemArray);
                if Succeeded(Result) then begin
                    //* Assign copy items task
                    Result := FileOp.CopyItems(ppsiItemArray, SiDestFolder);
                    if Succeeded(Result) then begin
                        //* Do the copy task
                        Result := FileOp.PerformOperations;
                    end;
                end;
            end;
        end;
        //OleCheck(Result);
    end;
end;

constructor TFOCopyFilesThread.Create(DestinationFolder: String; Files: TFOFileList; Handle: THandle);
begin
    inherited Create;
    FreeOnTerminate := True;
    FDestinationFolder := DestinationFolder;
    FFiles := Files;
    FHandle := Handle;
end;

constructor TFOCopyFilesThread.Create(DestinationFolder: PItemIDList; ItemIdList: TArrayOfPItemIDList; Handle: THandle);
begin
    inherited Create;
    FreeOnTerminate := True;
    FDestinationFolderPIDL := ILClone(DestinationFolder);
    CopyPIDLList(ItemIdList, FItemIdList);
    FHandle := Handle;
end;

destructor TFOCopyFilesThread.Destroy;
begin
    if Assigned(FItemIdList) then begin
        FreePIDLList(FItemIdList);
    end;
    if Assigned(FDestinationFolderPIDL) then begin
        CoTaskMemFree(FDestinationFolderPIDL);
    end;
    inherited;
end;

procedure TFOCopyFilesThread.Execute;
begin
    inherited;
    if Length(FFiles) > 0 then begin
        FOCopyFiles(FDestinationFolder, FFiles, FHandle);
    end else begin
        FOCopyFiles(FDestinationFolderPIDL, FItemIdList, FHandle);
    end;
end;

{ TFOMoveFilesThread }

function FOMoveFiles(DestinationFolder: String; Files: TFOFileList; Handle: THandle): HRESULT;
var
    i: Integer;
    FileOp: IFileOperation;
    SiDestFolder: IShellItem;
    ppsiItemArray: IShellItemArray;
    ItemIdList: TArrayOfPItemIDList;
    FileCount: Integer;
begin
    Result := S_FALSE;
    if Length(Files) = 0 then begin
        Exit;
    end;
    //create IFileOperation interface
    Result := CoCreateInstance(CLSID_FileOperation, nil, CLSCTX_ALL, IFileOperation, FileOp);
    if Succeeded(Result) then begin
        //set operations flags
        Result := FileOp.SetOperationFlags({FOF_NOCONFIRMATION OR} FOFX_NOMINIMIZEBOX);
        if Succeeded(Result) then begin
            //get destination folder shell item
            Result := SHCreateItemFromParsingName(PChar(DestinationFolder), nil, IShellItem, SiDestFolder);
            if Succeeded(Result) then begin
                FileCount := Length(Files);
                //* Our PIDL array
                SetLength(ItemIdList, Length(Files));
                for i := 0 to Length(Files) - 1 do begin
                    ItemIdList[i] := GetPIDLFromFileName(Files[i]);
                end;
                //* Create an array interface
                Result := SHCreateShellItemArrayFromIDLists(FileCount, PCUIDLIST_ABSOLUTE_ARRAY(ItemIdList), ppsiItemArray);
                if Succeeded(Result) then begin
                    //* Assign copy items task
                    Result := FileOp.MoveItems(ppsiItemArray, SiDestFolder);
                    if Succeeded(Result) then begin
                        //* Do the copy task
                        Result := FileOp.PerformOperations;
                    end;
                end;
                //* Free all source file PIDLs
                for i := 0 to Length(Files) - 1 do begin
                    CoTaskMemFree(ItemIdList[i]);
                end;
            end;
        end;
        //OleCheck(Result);
    end;
end;

function FOMoveFiles(DestinationFolder: PItemIDList; ItemIdList: TArrayOfPItemIDList; Handle: THandle): HRESULT;
var
    FileOp: IFileOperation;
    SiDestFolder: IShellItem;
    ppsiItemArray: IShellItemArray;
begin
    Result := S_FALSE;
    if (Length(ItemIdList) = 0)
    OR (NOT Assigned(DestinationFolder))
    OR (NOT Assigned(ItemIdList))
    then begin
        Exit;
    end;
    //create IFileOperation interface
    Result := CoCreateInstance(CLSID_FileOperation, nil, CLSCTX_ALL, IFileOperation, FileOp);
    if Succeeded(Result) then begin
        //set operations flags
        Result := FileOp.SetOperationFlags({FOF_NOCONFIRMATION OR} FOFX_NOMINIMIZEBOX);
        if Succeeded(Result) then begin
            //get destination folder shell item
            Result := SHCreateItemFromIDList(DestinationFolder, IShellItem, SiDestFolder);
            if Succeeded(Result) then begin
                //* Create an array interface
                Result := SHCreateShellItemArrayFromIDLists(Length(ItemIdList), PCUIDLIST_ABSOLUTE_ARRAY(ItemIdList), ppsiItemArray);
                if Succeeded(Result) then begin
                    //* Assign copy items task
                    Result := FileOp.MoveItems(ppsiItemArray, SiDestFolder);
                    if Succeeded(Result) then begin
                        //* Do the copy task
                        Result := FileOp.PerformOperations;
                    end;
                end;
            end;
        end;
        //OleCheck(Result);
    end;
end;

constructor TFOMoveFilesThread.Create(DestinationFolder: String; Files: TFOFileList; Handle: THandle);
begin
    inherited Create;
    FreeOnTerminate := True;
    FDestinationFolder := DestinationFolder;
    FFiles := Files;
    FHandle := Handle;
end;

constructor TFOMoveFilesThread.Create(DestinationFolder: PItemIDList; ItemIdList: TArrayOfPItemIDList; Handle: THandle);
begin
    inherited Create;
    FreeOnTerminate := True;
    FDestinationFolderPIDL := ILClone(DestinationFolder);
    CopyPIDLList(ItemIdList, FItemIdList);
    FHandle := Handle;
end;

destructor TFOMoveFilesThread.Destroy;
begin
    if Assigned(FItemIdList) then begin
        FreePIDLList(FItemIdList);
    end;
    if Assigned(FDestinationFolderPIDL) then begin
        CoTaskMemFree(FDestinationFolderPIDL);
    end;
    inherited;
end;

procedure TFOMoveFilesThread.Execute;
begin
    inherited;
    if Length(FFiles) > 0 then begin
        FOMoveFiles(FDestinationFolder, FFiles, FHandle);
    end else begin
        FOMoveFiles(FDestinationFolderPIDL, FItemIdList, FHandle);
    end;
end;

{ TFODeleteFilesThread }

constructor TFODeleteFilesThread.Create(Files: TFOFileList; Permanently: Boolean; Handle: THandle);
begin
    inherited Create;
    FreeOnTerminate := True;
    FFiles := Files;
    FPermanently := Permanently;
    FHandle := Handle;
end;

constructor TFODeleteFilesThread.Create(ItemIdList: TArrayOfPItemIDList; Permanently: Boolean; Handle: THandle);
begin
    inherited Create;
    FreeOnTerminate := True;
    CopyPIDLList(ItemIdList, FItemIdList);
    FPermanently := Permanently;
    FHandle := Handle;
end;

destructor TFODeleteFilesThread.Destroy;
begin
    if Assigned(FItemIdList) then begin
        FreePIDLList(FItemIdList);
    end;
    inherited;
end;

procedure TFODeleteFilesThread.Execute;
begin
    inherited;
    if Length(FFiles) > 0 then begin
        FODeleteFiles(FFiles, FPermanently, FHandle);
    end else begin
        FODeleteFiles(FItemIdList, FPermanently, FHandle);
    end;
end;

function FOCopyFile(DestinationFolder: String; FileName: String; Handle: THandle): HRESULT;
var
    Files: TFOFileList;
begin
    SetLength(Files, 1);
    Files[0] := FileName;
    Result := FOCopyFiles(DestinationFolder, Files, Handle);
end;

function FOCopyFile(DestinationFolder: PItemIDList; Item: PItemIDList; Handle: THandle): HRESULT;
var
    ItemIDList: TArrayOfPItemIDList;
begin
    SetLength(ItemIDList, 1);
    ItemIDList[0] := Item;
    Result := FOCopyFiles(DestinationFolder, ItemIDList, Handle);
end;

function FOMoveFile(DestinationFolder: String; FileName: String; Handle: THandle): HRESULT;
var
    Files: TFOFileList;
begin
    SetLength(Files, 1);
    Files[0] := FileName;
    Result := FOMoveFiles(DestinationFolder, Files, Handle);
end;

function FOMoveFile(DestinationFolder: PItemIDList; Item: PItemIDList; Handle: THandle): HRESULT;
var
    ItemIDList: TArrayOfPItemIDList;
begin
    SetLength(ItemIDList, 1);
    ItemIDList[0] := Item;
    Result := FOMoveFiles(DestinationFolder, ItemIDList, Handle);
end;

function FODeleteFile(FileItem: String; Permanently: Boolean; Handle: THandle): HRESULT; overload;
var
    FileOp: IFileOperation;
    Item: IShellItem;
    Flags: Cardinal;
begin
    Result := S_FALSE;
    if (NOT FileExists(FileItem))
    AND (NOT DirectoryExists(FileItem))
    then begin
        Exit;
    end;
    //create IFileOperation interface
    Result := CoCreateInstance(CLSID_FileOperation, nil, CLSCTX_ALL, IFileOperation, FileOp);
    if Succeeded(Result) then begin
        //set operations flags
        Flags := FOFX_NOMINIMIZEBOX;
        if NOT Permanently then begin
            Flags := Flags OR FOF_ALLOWUNDO;
        end;
        Result := FileOp.SetOperationFlags(Flags);
        if Succeeded(Result) then begin
            //* Create an array interface
            Result := SHCreateItemFromParsingName(PChar(FileItem), nil, IShellItem, Item);
            if Succeeded(Result) then begin
                //* Assign copy items task
                Result := FileOp.DeleteItem(Item, nil);
                if Succeeded(Result) then begin
                    //* Do the copy task
                    Result := FileOp.PerformOperations;
                end;
            end;
        end;
        //OleCheck(Result);
    end;
end;

function FODeleteFile(Item: PItemIDList; Permanently: Boolean; Handle: THandle): HRESULT; overload;
var
    FileOp: IFileOperation;
    ShellItem: IShellItem;
    Flags: Cardinal;
begin
    Result := S_FALSE;
    if NOT Assigned(Item) then begin
        Exit;
    end;
    //create IFileOperation interface
    Result := CoCreateInstance(CLSID_FileOperation, nil, CLSCTX_ALL, IFileOperation, FileOp);
    if Succeeded(Result) then begin
        //set operations flags
        Flags := FOFX_NOMINIMIZEBOX;
        if NOT Permanently then begin
            Flags := Flags OR FOF_ALLOWUNDO;
        end;
        Result := FileOp.SetOperationFlags(Flags);
        if Succeeded(Result) then begin
            //* Create an array interface
            Result := SHCreateItemFromIDList(Item, IShellItem, ShellItem);
            if Succeeded(Result) then begin
                //* Assign copy items task
                Result := FileOp.DeleteItem(ShellItem, nil);
                if Succeeded(Result) then begin
                    //* Do the copy task
                    Result := FileOp.PerformOperations;
                end;
            end;
        end;
        //OleCheck(Result);
    end;
end;

function FODeleteFiles(Files: TFOFileList; Permanently: Boolean; Handle: THandle): HRESULT;
var
    i: Integer;
    FileOp: IFileOperation;
    ppsiItemArray: IShellItemArray;
    ItemIdList: TArrayOfPItemIDList;
    FileCount: Integer;
    Flags: Cardinal;
begin
    Result := S_FALSE;
    if Length(Files) = 0 then begin
        Exit;
    end;
    //create IFileOperation interface
    Result := CoCreateInstance(CLSID_FileOperation, nil, CLSCTX_ALL, IFileOperation, FileOp);
    if Succeeded(Result) then begin
        //set operations flags
        Flags := FOFX_NOMINIMIZEBOX;
        if NOT Permanently then begin
            Flags := Flags OR FOF_ALLOWUNDO;
        end;
        Result := FileOp.SetOperationFlags(Flags);
        if Succeeded(Result) then begin
            FileCount := Length(Files);
            //* Our PIDL array
            SetLength(ItemIdList, Length(Files));
            for i := 0 to Length(Files) - 1 do begin
                ItemIdList[i] := GetPIDLFromFileName(Files[i]);
            end;
            //* Create an array interface
            Result := SHCreateShellItemArrayFromIDLists(FileCount, PCUIDLIST_ABSOLUTE_ARRAY(ItemIdList), ppsiItemArray);
            if Succeeded(Result) then begin
                //* Assign copy items task
                Result := FileOp.DeleteItems(ppsiItemArray);
                if Succeeded(Result) then begin
                    //* Do the copy task
                    Result := FileOp.PerformOperations;
                end;
            end;
            //* Free all source file PIDLs
            for i := 0 to Length(Files) - 1 do begin
                CoTaskMemFree(ItemIdList[i]);
            end;
        end;
        //OleCheck(Result);
    end;
end;

function FODeleteFiles(ItemIdList: TArrayOfPItemIDList; Permanently: Boolean; Handle: THandle): HRESULT;
var
    FileOp: IFileOperation;
    ppsiItemArray: IShellItemArray;
    Flags: Cardinal;
begin
    Result := S_FALSE;
    if (Length(ItemIdList) = 0)
    OR (NOT Assigned(ItemIdList))
    then begin
        Exit;
    end;
    //create IFileOperation interface
    Result := CoCreateInstance(CLSID_FileOperation, nil, CLSCTX_ALL, IFileOperation, FileOp);
    if Succeeded(Result) then begin
        //set operations flags
        Flags := FOFX_NOMINIMIZEBOX;
        if NOT Permanently then begin
            Flags := Flags OR FOF_ALLOWUNDO;
        end;
        Result := FileOp.SetOperationFlags(Flags);
        if Succeeded(Result) then begin
            //* Create an array interface
            Result := SHCreateShellItemArrayFromIDLists(Length(ItemIdList), PCUIDLIST_ABSOLUTE_ARRAY(ItemIdList), ppsiItemArray);
            if Succeeded(Result) then begin
                //* Assign copy items task
                Result := FileOp.DeleteItems(ppsiItemArray);
                if Succeeded(Result) then begin
                    //* Do the copy task
                    Result := FileOp.PerformOperations;
                end;
            end;
        end;
        //OleCheck(Result);
    end;
end;

function FOCopyFileAs(DestinationFolder: String; FileName: String; NewFileName: String; Handle: THandle): HRESULT;
var
    FileOp: IFileOperation;
    SiDestFolder: IShellItem;
    Item: IShellItem;
begin
    Result := S_FALSE;
    if (FileName = '')
    OR (DestinationFolder = '')
    then begin
        Exit;
    end;
    //create IFileOperation interface
    Result := CoCreateInstance(CLSID_FileOperation, nil, CLSCTX_ALL, IFileOperation, FileOp);
    if Succeeded(Result) then begin
        //set operations flags
        Result := FileOp.SetOperationFlags({FOF_NOCONFIRMATION OR} FOFX_NOMINIMIZEBOX);
        if Succeeded(Result) then begin
            //get destination folder shell item
            Result := SHCreateItemFromParsingName(PChar(DestinationFolder), nil, IShellItem, SiDestFolder);
            if Succeeded(Result) then begin
                //* Our PIDL array
                Result := SHCreateItemFromParsingName(PChar(FileName), nil, IShellItem, Item);
                if Succeeded(Result) then begin
                    //* Assign copy items task
                    Result := FileOp.CopyItem(Item, SiDestFolder, PChar(NewFileName), nil);
                    if Succeeded(Result) then begin
                        //* Do the copy task
                        Result := FileOp.PerformOperations;
                    end;
                end;
            end;
        end;
        //OleCheck(Result);
    end;
end;

function FOCopyFileAs(DestinationFolder: PItemIDList; FileName: PItemIDList; NewFileName: String; Handle: THandle): HRESULT;
var
    FileOp: IFileOperation;
    SiDestFolder: IShellItem;
    Item: IShellItem;
begin
    Result := S_FALSE;
    if (NewFileName = '')
    OR (NOT Assigned(DestinationFolder))
    OR (NOT Assigned(FileName))
    then begin
        Exit;
    end;
    //create IFileOperation interface
    Result := CoCreateInstance(CLSID_FileOperation, nil, CLSCTX_ALL, IFileOperation, FileOp);
    if Succeeded(Result) then begin
        //set operations flags
        Result := FileOp.SetOperationFlags({FOF_NOCONFIRMATION OR} FOFX_NOMINIMIZEBOX);
        if Succeeded(Result) then begin
            //get destination folder shell item
            Result := SHCreateItemFromIDList(DestinationFolder, IShellItem, SiDestFolder);
            if Succeeded(Result) then begin
                //* Our PIDL array
                Result := SHCreateItemFromIDList(FileName, IShellItem, Item);
                if Succeeded(Result) then begin
                    //* Assign copy items task
                    Result := FileOp.CopyItem(Item, SiDestFolder, PChar(NewFileName), nil);
                    if Succeeded(Result) then begin
                        //* Do the copy task
                        Result := FileOp.PerformOperations;
                    end;
                end;
            end;
        end;
        //OleCheck(Result);
    end;
end;

function FOMoveFileAs(DestinationFolder: String; FileName: String; NewFileName: String; Handle: THandle): HRESULT;
var
    FileOp: IFileOperation;
    SiDestFolder: IShellItem;
    Item: IShellItem;
begin
    Result := S_FALSE;
    if (FileName = '')
    OR (DestinationFolder = '')
    then begin
        Exit;
    end;
    //create IFileOperation interface
    Result := CoCreateInstance(CLSID_FileOperation, nil, CLSCTX_ALL, IFileOperation, FileOp);
    if Succeeded(Result) then begin
        //set operations flags
        Result := FileOp.SetOperationFlags({FOF_NOCONFIRMATION OR} FOFX_NOMINIMIZEBOX);
        if Succeeded(Result) then begin
            //get destination folder shell item
            Result := SHCreateItemFromParsingName(PChar(DestinationFolder), nil, IShellItem, SiDestFolder);
            if Succeeded(Result) then begin
                Result := SHCreateItemFromParsingName(PChar(FileName), nil, IShellItem, Item);
                //* Create an array interface
                if Succeeded(Result) then begin
                    //* Assign copy items task
                    Result := FileOp.MoveItem(Item, SiDestFolder, PChar(NewFileName), nil);
                    if Succeeded(Result) then begin
                        //* Do the copy task
                        Result := FileOp.PerformOperations;
                    end;
                end;
            end;
        end;
        //OleCheck(Result);
    end;
end;

function FOMoveFileAs(DestinationFolder: PItemIDList; FileName: PItemIDList; NewFileName: String; Handle: THandle): HRESULT;
var
    FileOp: IFileOperation;
    SiDestFolder: IShellItem;
    Item: IShellItem;
begin
    Result := S_FALSE;
    if (NewFileName = '')
    OR (NOT Assigned(DestinationFolder))
    OR (NOT Assigned(FileName))
    then begin
        Exit;
    end;
    //create IFileOperation interface
    Result := CoCreateInstance(CLSID_FileOperation, nil, CLSCTX_ALL, IFileOperation, FileOp);
    if Succeeded(Result) then begin
        //set operations flags
        Result := FileOp.SetOperationFlags({FOF_NOCONFIRMATION OR} FOFX_NOMINIMIZEBOX);
        if Succeeded(Result) then begin
            //get destination folder shell item
            Result := SHCreateItemFromIDList(DestinationFolder, IShellItem, SiDestFolder);
            if Succeeded(Result) then begin
                Result := SHCreateItemFromIDList(FileName, IShellItem, Item);
                //* Create an array interface
                if Succeeded(Result) then begin
                    //* Assign copy items task
                    Result := FileOp.MoveItem(Item, SiDestFolder, PChar(NewFileName), nil);
                    if Succeeded(Result) then begin
                        //* Do the copy task
                        Result := FileOp.PerformOperations;
                    end;
                end;
            end;
        end;
        //OleCheck(Result);
    end;
end;

function ZIPFolderExists(ZipFile: TZipFile; Path: String): Boolean;
var
    ItemName: String;
    i: Integer;
begin
    Result := False;
    for i := 0 to Length(ZipFile.FileNames) - 1 do begin
        ItemName := Copy(ZipFile.FileNames[i], 0, Length(Path));
        if SameText(ItemName, Path) then begin
            Result := True;
            Break;
        end;
    end;
end;

function FOCreateFolder(DestinationFolder: PItemIDList; FolderName: String; Handle: THandle; var IsZIP: Boolean): HRESULT; overload;
var
    IsZIPFolderDestination: Boolean;
    ZIPFileName: String;
    ZIPSubFolder: String;
    ZipFile: TZipFile;
    FileNameInZIP: String;
    //Bytes: TBytes;
    FileOp: IFileOperation;
    SiDestFolder: IShellItem;
begin
    Result := S_FALSE;
    IsZIP := False;
    if (FolderName = '')
    OR (NOT Assigned(DestinationFolder))
    then begin
        Exit;
    end;
    //* Check if destination is a ZIP folder (or ZIP sub-folder)
    IsZIPFolderDestination := IsZIPFolder(DestinationFolder, ZIPFileName);
    if IsZIPFolderDestination then begin
        IsZIP := True;
        ZIPSubFolder := ZIPFolderPath(DestinationFolder);
        ZipFile := TZipFile.Create;
        try
            ZipFile.Open(ZIPFileName, zmReadWrite);
            if ZIPSubFolder = '' then begin
                FileNameInZIP := FolderName + '/';
            end else begin
                FileNameInZIP := ZIPSubFolder + '/' + FolderName + '/';
            end;

            FileNameInZIP := StringReplace(FileNameInZIP, '\', '/', [rfReplaceAll]);

            if ZipFile.IndexOf(FileNameInZIP) > - 1 then begin
                Exit;
            end;

            if ZIPFolderExists(ZipFile, FileNameInZIP) then begin
                Exit;
            end;

            //ZipFile.Add(Bytes, FileNameInZIP);

            ZipFile.CheckToCreatePath(FileNameInZIP);

            Result := S_OK;
        finally
            ZipFile.Close;
            FreeAndNil(ZipFile);
        end;
        Exit;
    end;
    //create IFileOperation interface
    Result := CoCreateInstance(CLSID_FileOperation, nil, CLSCTX_ALL, IFileOperation, FileOp);
    if Succeeded(Result) then begin
        //set operations flags
        Result := FileOp.SetOperationFlags({FOF_NOCONFIRMATION OR} FOFX_NOMINIMIZEBOX);
        if Succeeded(Result) then begin
            //get destination folder shell item
            Result := SHCreateItemFromIDList(DestinationFolder, IShellItem, SiDestFolder);
            if Succeeded(Result) then begin
                //* Assign copy items task
                Result := FileOp.NewItem(SiDestFolder, FILE_ATTRIBUTE_DIRECTORY, PChar(FolderName), nil, nil);
                if Succeeded(Result) then begin
                    //* Do the copy task
                    Result := FileOp.PerformOperations;
                end;
            end;
        end;
        //OleCheck(Result);
    end;
end;

function IsZIPFolder(PIDL: PItemIDList): Boolean;
var
    ZIPFileName: String;
begin
    Result := IsZIPFolder(PIDL, ZIPFileName);
end;

function IsZIPFolder(PIDL: PItemIDList; var ZIPFileName: String): Boolean;
var
    FileName: String;
    WorkPIDL: PItemIDList;
begin
    Result := False;
    if NOT Assigned(PIDL) then begin
        Exit;
    end;
    FileName := GetFileNameFromPIDL(PIDL);
    if FileExists(FileName)
    AND FileName.EndsWith('.ZIP', True)
    then begin
        ZIPFileName := FileName;
        Result := True;
        Exit;
    end;
    if (FileName <> '')
    AND DirectoryExists(FileName)
    then begin
        Exit;
    end;
    WorkPIDL := ILClone(PIDL);
    try
        while ILRemoveLastID(WorkPIDL) do begin
            FileName := GetFileNameFromPIDL(WorkPIDL);
            if (FileName <> '')
            AND FileExists(FileName)
            then begin
                Result := True;
                ZIPFileName := FileName;
                Break;
            end;
        end;
    finally
        CoTaskMemFree(WorkPIDL);
    end;
end;

function ZIPFolderPath(PIDL: PItemIDList): String;
var
    FileName: String;
    WorkPIDL: PItemIDList;
begin
    Result := '';
    FileName := GetFileNameFromPIDL(PIDL);
    if FileExists(FileName)
    AND FileName.EndsWith('.ZIP', True)
    then begin
        Result := '';
        Exit;
    end;
    WorkPIDL := ILClone(PIDL);
    try
        Result := GetPIDLDisplayName(WorkPIDL);
        while ILRemoveLastID(WorkPIDL) do begin
            FileName := GetFileNameFromPIDL(WorkPIDL);
            if (FileName <> '')
            AND FileExists(FileName)
            then begin
                Break;
            end else begin
                Result := GetPIDLDisplayName(WorkPIDL) + '/' + Result;
            end;
        end;
    finally
        CoTaskMemFree(WorkPIDL);
    end;
end;

function GetSourceFileListFileSystem(ItemIdList: TArrayOfPItemIDList): TStringList;
var
    i: Integer;
    FileName: String;
begin
    Result := TStringList.Create;
    for i := 0 to High(ItemIdList) do begin
        FileName := GetFileNameFromPIDL(ItemIdList[i]);
        if FileExists(FileName)
        OR DirectoryExists(FileName)
        then begin
            Result.Append(FileName);
        end;
    end;
end;

function FOCopyFilesToZIP(DestinationFolder: PItemIDList; ItemIdList: TArrayOfPItemIDList; Handle: THandle; ZipProgressEvent: TZipProgressEvent = nil; ZIPProgressCompleted: TZIPProgressCompleted = nil): HRESULT; overload;
var
    IsZIPFolderDestination: Boolean;
    ZIPFileName: String;
    SourceFolderPath: String;
    ZIPSubFolder: String;
    ZipFile: TZipFile;
    SourceFileListFileSystem: TStringList;
    i, k: Integer;
    FileNameInZIP: String;
    FolderFiles: TStringDynArray;
    Asked: Boolean;
    YesCurrent: Boolean;
    YesToAll: Boolean;
begin
    Result := S_FALSE;
    if (NOT Assigned(DestinationFolder))
    OR (NOT Assigned(ItemIdList))
    OR (Length(ItemIdList) = 0)
    then begin
        Exit;
    end;
    Asked := False;
    YesCurrent := False;
    YesToAll := False;
    IsZIPFolderDestination := IsZIPFolder(DestinationFolder, ZIPFileName);
    if NOT IsZIPFolderDestination then begin
        Exit;
    end;
    ZIPSubFolder := ZIPFolderPath(DestinationFolder);
    SourceFileListFileSystem := GetSourceFileListFileSystem(ItemIdList);
    try
        if SourceFileListFileSystem.Count > 0 then begin
            ZipFile := TZipFile.Create;
            try
                ZipFile.OnProgress := ZipProgressEvent;
                ZipFile.Open(ZIPFileName, zmReadWrite);
                for i := 0 to SourceFileListFileSystem.Count - 1 do begin

                    SourceFolderPath := ExcludeTrailingPathDelimiter(ExtractFilePath(SourceFileListFileSystem[i]));

                    //* Single file
                    if FileExists(SourceFileListFileSystem[i]) then begin
                        if ZIPSubFolder = '' then begin
                            FileNameInZIP := Copy(SourceFileListFileSystem[i], Length(SourceFolderPath) + 2, Length(SourceFileListFileSystem[i]));
                        end else begin
                            FileNameInZIP := ZIPSubFolder + '\' + Copy(SourceFileListFileSystem[i], Length(SourceFolderPath) + 2, Length(SourceFileListFileSystem[i]));
                        end;

                        FileNameInZIP := StringReplace(FileNameInZIP, '\', '/', [rfReplaceAll]);

                        if ZipFile.IndexOf(FileNameInZIP) > - 1 then begin
                            if NOT Asked then begin
                                YesCurrent := False;
                                case MessageDlg('File already exists in ZIP: ' + FileNameInZIP + #13#10 + 'Overwrite?', mtConfirmation, [mbYes, mbNo, mbCancel, mbNoToAll, mbYesToAll], 0) of
                                    mrYes: YesCurrent := True;
                                    mrNo: Continue;
                                    mrCancel: begin
                                        Result := S_OK;
                                        Exit;
                                    end;
                                    mrNoToAll: begin
                                        Asked := True;
                                        YesToAll := False;
                                        Continue;
                                    end;
                                    mrYesToAll: begin
                                        Asked := True;
                                        YesToAll := True;
                                    end;
                                end;
                            end;
                            if YesCurrent
                            OR YesToAll
                            then begin
                                ZipFile.Delete(FileNameInZIP);
                            end else begin
                                Continue;
                            end;
                        end;

                        if ZipFile.IndexOf(ExtractFilePathZIP(FileNameInZIP)) > - 1 then begin
                            ZipFile.Delete(ExtractFilePathZIP(FileNameInZIP));
                        end;

                        ZipFile.Add(SourceFileListFileSystem[i], FileNameInZIP);
                        if Assigned(ZIPProgressCompleted) then begin
                            ZIPProgressCompleted(SourceFileListFileSystem[i], NGetFileSize(SourceFileListFileSystem[i]));
                        end;
                    //* Files in sub-folders
                    end else if DirectoryExists(SourceFileListFileSystem[i]) then begin
                        FolderFiles := TDirectory.GetFiles(SourceFileListFileSystem[i], '*', TSearchOption.soAllDirectories);
                        for k := Low(FolderFiles) to High(FolderFiles) do begin
                            if ZIPSubFolder = '' then begin
                                FileNameInZIP := Copy(FolderFiles[k], Length(SourceFolderPath) + 2, Length(FolderFiles[k]));
                            end else begin
                                FileNameInZIP := ZIPSubFolder + '/' + Copy(FolderFiles[k], Length(SourceFolderPath) + 2, Length(FolderFiles[k]));
                            end;

                            FileNameInZIP := StringReplace(FileNameInZIP, '\', '/', [rfReplaceAll]);

                            if FileExists(FolderFiles[k]) then begin
                                if ZipFile.IndexOf(FileNameInZIP) > - 1 then begin
                                    if NOT Asked then begin
                                        YesCurrent := False;
                                        case MessageDlg('File already exists in ZIP: ' + FileNameInZIP + #13#10 + 'Overwrite?', mtConfirmation, [mbYes, mbNo, mbCancel, mbNoToAll, mbYesToAll], 0) of
                                            mrYes: YesCurrent := True;
                                            mrNo: Continue;
                                            mrCancel: begin
                                                Result := S_OK;
                                                Exit;
                                            end;
                                            mrNoToAll: begin
                                                Asked := True;
                                                YesToAll := False;
                                                Continue;
                                            end;
                                            mrYesToAll: begin
                                                Asked := True;
                                                YesToAll := True;
                                            end;
                                        end;
                                    end;
                                    if YesCurrent
                                    OR YesToAll
                                    then begin
                                        ZipFile.Delete(FileNameInZIP);
                                    end else begin
                                        Continue;
                                    end;
                                end;
                            end;

                            if ZipFile.IndexOf(ExtractFilePathZIP(FileNameInZIP)) > - 1 then begin
                                ZipFile.Delete(ExtractFilePathZIP(FileNameInZIP));
                            end;

                            ZipFile.Add(FolderFiles[k], FileNameInZIP);
                            if Assigned(ZIPProgressCompleted) then begin
                                ZIPProgressCompleted(FolderFiles[k], NGetFileSize(FolderFiles[k]));
                            end;
                        end;
                    end;
                end;
                Result := S_OK;
            finally
                ZipFile.Close;
                FreeAndNil(ZipFile);
            end;
        end;
    finally
        FreeAndNil(SourceFileListFileSystem);
    end;
end;

function FOMoveFilesToZIP(DestinationFolder: PItemIDList; ItemIdList: TArrayOfPItemIDList; Handle: THandle; ZipProgressEvent: TZipProgressEvent = nil; ZIPProgressCompleted: TZIPProgressCompleted = nil): HRESULT; overload;
var
    IsZIPFolderDestination: Boolean;
    ZIPFileName: String;
    SourceFolderPath: String;
    ZIPSubFolder: String;
    ZipFile: TZipFile;
    SourceFileListFileSystem: TStringList;
    i, k: Integer;
    FileNameInZIP: String;
    FolderFiles: TStringDynArray;
    Asked: Boolean;
    YesCurrent: Boolean;
    YesToAll: Boolean;
    FolderNotCompletelyCopied: Boolean;
begin
    {
    Result := FOCopyFilesToZIP(DestinationFolder, ItemIdList, Handle, ZipProgressEvent, ZIPProgressCompleted);
    if Result = S_OK then begin
        if IsZIPFolder(ItemIdList[0], ZIPFileName) then begin
            Result := FODeleteFilesInZIP(ItemIdList, Handle, ZipProgressEvent, ZIPProgressCompleted);
        end else begin
            Result := FODeleteFiles(ItemIdList, False, Handle);
        end;
    end;
    }
    Result := S_FALSE;
    if (NOT Assigned(DestinationFolder))
    OR (NOT Assigned(ItemIdList))
    OR (Length(ItemIdList) = 0)
    then begin
        Exit;
    end;
    Asked := False;
    YesCurrent := False;
    YesToAll := False;
    IsZIPFolderDestination := IsZIPFolder(DestinationFolder, ZIPFileName);
    if NOT IsZIPFolderDestination then begin
        Exit;
    end;
    ZIPSubFolder := ZIPFolderPath(DestinationFolder);
    SourceFileListFileSystem := GetSourceFileListFileSystem(ItemIdList);
    try
        if SourceFileListFileSystem.Count > 0 then begin
            ZipFile := TZipFile.Create;
            try
                ZipFile.OnProgress := ZipProgressEvent;
                ZipFile.Open(ZIPFileName, zmReadWrite);
                for i := 0 to SourceFileListFileSystem.Count - 1 do begin

                    SourceFolderPath := ExcludeTrailingPathDelimiter(ExtractFilePath(SourceFileListFileSystem[i]));

                    //* Single file
                    if FileExists(SourceFileListFileSystem[i]) then begin
                        if ZIPSubFolder = '' then begin
                            FileNameInZIP := Copy(SourceFileListFileSystem[i], Length(SourceFolderPath) + 2, Length(SourceFileListFileSystem[i]));
                        end else begin
                            FileNameInZIP := ZIPSubFolder + '\' + Copy(SourceFileListFileSystem[i], Length(SourceFolderPath) + 2, Length(SourceFileListFileSystem[i]));
                        end;

                        FileNameInZIP := StringReplace(FileNameInZIP, '\', '/', [rfReplaceAll]);

                        if ZipFile.IndexOf(FileNameInZIP) > - 1 then begin
                            if NOT Asked then begin
                                YesCurrent := False;
                                case MessageDlg('File already exists in ZIP: ' + FileNameInZIP + #13#10 + 'Overwrite?', mtConfirmation, [mbYes, mbNo, mbCancel, mbNoToAll, mbYesToAll], 0) of
                                    mrYes: YesCurrent := True;
                                    mrNo: Continue;
                                    mrCancel: begin
                                        Result := S_OK;
                                        Exit;
                                    end;
                                    mrNoToAll: begin
                                        Asked := True;
                                        YesToAll := False;
                                        Continue;
                                    end;
                                    mrYesToAll: begin
                                        Asked := True;
                                        YesToAll := True;
                                    end;
                                end;
                            end;
                            if YesCurrent
                            OR YesToAll
                            then begin
                                ZipFile.Delete(FileNameInZIP);
                            end else begin
                                Continue;
                            end;
                        end;

                        if ZipFile.IndexOf(ExtractFilePathZIP(FileNameInZIP)) > - 1 then begin
                            ZipFile.Delete(ExtractFilePathZIP(FileNameInZIP));
                        end;

                        ZipFile.Add(SourceFileListFileSystem[i], FileNameInZIP);

                        SysUtils.DeleteFile(SourceFileListFileSystem[i]);

                        if Assigned(ZIPProgressCompleted) then begin
                            ZIPProgressCompleted(SourceFileListFileSystem[i], NGetFileSize(SourceFileListFileSystem[i]));
                        end;
                    //* Files in sub-folders
                    end else if DirectoryExists(SourceFileListFileSystem[i]) then begin
                        FolderFiles := TDirectory.GetFiles(SourceFileListFileSystem[i], '*', TSearchOption.soAllDirectories);
                        FolderNotCompletelyCopied := False;
                        for k := Low(FolderFiles) to High(FolderFiles) do begin
                            if ZIPSubFolder = '' then begin
                                FileNameInZIP := Copy(FolderFiles[k], Length(SourceFolderPath) + 2, Length(FolderFiles[k]));
                            end else begin
                                FileNameInZIP := ZIPSubFolder + '/' + Copy(FolderFiles[k], Length(SourceFolderPath) + 2, Length(FolderFiles[k]));
                            end;

                            FileNameInZIP := StringReplace(FileNameInZIP, '\', '/', [rfReplaceAll]);

                            if FileExists(FolderFiles[k]) then begin
                                if ZipFile.IndexOf(FileNameInZIP) > - 1 then begin
                                    if NOT Asked then begin
                                        YesCurrent := False;
                                        case MessageDlg('File already exists in ZIP: ' + FileNameInZIP + #13#10 + 'Overwrite?', mtConfirmation, [mbYes, mbNo, mbCancel, mbNoToAll, mbYesToAll], 0) of
                                            mrYes: YesCurrent := True;
                                            mrNo: begin
                                                FolderNotCompletelyCopied := True;
                                                Continue;
                                            end;
                                            mrCancel: begin
                                                Result := S_OK;
                                                Exit;
                                            end;
                                            mrNoToAll: begin
                                                Asked := True;
                                                YesToAll := False;
                                                FolderNotCompletelyCopied := True;
                                                Continue;
                                            end;
                                            mrYesToAll: begin
                                                Asked := True;
                                                YesToAll := True;
                                            end;
                                        end;
                                    end;
                                    if YesCurrent
                                    OR YesToAll
                                    then begin
                                        ZipFile.Delete(FileNameInZIP);
                                    end else begin
                                        Continue;
                                    end;
                                end;
                            end;

                            if ZipFile.IndexOf(ExtractFilePathZIP(FileNameInZIP)) > - 1 then begin
                                ZipFile.Delete(ExtractFilePathZIP(FileNameInZIP));
                            end;

                            ZipFile.Add(FolderFiles[k], FileNameInZIP);

                            SysUtils.DeleteFile(FolderFiles[k]);

                            if Assigned(ZIPProgressCompleted) then begin
                                ZIPProgressCompleted(FolderFiles[k], NGetFileSize(FolderFiles[k]));
                            end;
                        end;
                        if NOT FolderNotCompletelyCopied then begin
                            FODeleteFile(SourceFileListFileSystem[i], False, Handle);
                        end;
                    end;
                end;
                Result := S_OK;
            finally
                ZipFile.Close;
                FreeAndNil(ZipFile);
            end;
        end;
    finally
        FreeAndNil(SourceFileListFileSystem);
    end;
end;

function FOCopyFileToZIPAs(DestinationFolder: PItemIDList; Item: PItemIDList; NewFileName: String; Handle: THandle; ZipProgressEvent: TZipProgressEvent = nil; ZIPProgressCompleted: TZIPProgressCompleted = nil): HRESULT; overload;
var
    IsZIPFolderDestination: Boolean;
    ZIPFileName: String;
    SourceFolderPath: String;
    ZIPSubFolder: String;
    ZipFile: TZipFile;
    SourceFile: String;
    NewFile: String;
    NewPath: String;
    k: Integer;
    FileNameInZIP: String;
    FolderFiles: TStringDynArray;
    //Asked: Boolean;
    //YesCurrent: Boolean;
    //YesToAll: Boolean;
begin
    Result := S_FALSE;
    if (NOT Assigned(DestinationFolder))
    OR (NOT Assigned(Item))
    then begin
        Exit;
    end;
    //Asked := False;
    //YesCurrent := False;
    //YesToAll := False;
    IsZIPFolderDestination := IsZIPFolder(DestinationFolder, ZIPFileName);
    if NOT IsZIPFolderDestination then begin
        Exit;
    end;
    ZIPSubFolder := ZIPFolderPath(DestinationFolder);
    SourceFile := GetFileNameFromPIDL(Item);
    ZipFile := TZipFile.Create;
    try

        SourceFolderPath := ExcludeTrailingPathDelimiter(ExtractFilePath(SourceFile));

        ZipFile.OnProgress := ZipProgressEvent;
        ZipFile.Open(ZIPFileName, zmReadWrite);
        //* Single file
        if FileExists(SourceFile) then begin
            NewFile := IncludeTrailingPathDelimiter(ExtractFilePath(SourceFile)) + NewFileName;
            if ZIPSubFolder = '' then begin
                FileNameInZIP := Copy(NewFile, Length(SourceFolderPath) + 2, Length(NewFile));
            end else begin
                FileNameInZIP := ZIPSubFolder + '/' + Copy(NewFile, Length(SourceFolderPath) + 2, Length(NewFile));
            end;
            FileNameInZIP := StringReplace(FileNameInZIP, '\', '/', [rfReplaceAll]);

            if ZipFile.IndexOf(FileNameInZIP) > - 1 then begin
                //if NOT Asked then begin
                    //YesCurrent := False;
                    case MessageDlg('File already exists in ZIP: ' + FileNameInZIP + #13#10 + 'Overwrite?', mtConfirmation, [mbYes, mbNo, mbCancel{, mbNoToAll, mbYesToAll}], 0) of
                        //mrYes: YesCurrent := True;
                        mrNo: begin
                            Result := S_OK;
                            Exit;
                        end;
                        mrCancel: begin
                            Result := S_OK;
                            Exit;
                        end;
                        {
                        mrNoToAll: begin
                            Asked := True;
                            YesToAll := False;
                        end;
                        mrYesToAll: begin
                            Asked := True;
                            YesToAll := True;
                        end;
                        }
                    end;
                //end;
                //if YesCurrent
                //OR YesToAll
                //then begin
                    ZipFile.Delete(FileNameInZIP);
                //end;
            end;

            if ZipFile.IndexOf(ExtractFilePathZIP(FileNameInZIP)) > - 1 then begin
                ZipFile.Delete(ExtractFilePathZIP(FileNameInZIP));
            end;

            ZipFile.Add(SourceFile, FileNameInZIP);
            if Assigned(ZIPProgressCompleted) then begin
                ZIPProgressCompleted(SourceFile, NGetFileSize(SourceFile));
            end;
            Result := S_OK;
        //* Files in sub-folders
        end else if DirectoryExists(SourceFile) then begin
            FolderFiles := TDirectory.GetFiles(SourceFile, '*', TSearchOption.soAllDirectories);
            for k := Low(FolderFiles) to High(FolderFiles) do begin
                NewPath := Copy(FolderFiles[k], Length(SourceFolderPath) + 2, Length(FolderFiles[k]));
                NewPath := Copy(NewPath, Pos('/', NewPath), Length(NewPath));
                NewPath := NewFileName + NewPath;
                if ZIPSubFolder = '' then begin
                    FileNameInZIP := NewPath;
                end else begin
                    FileNameInZIP := ZIPSubFolder + '/' + NewPath;
                end;

                FileNameInZIP := StringReplace(FileNameInZIP, '\', '/', [rfReplaceAll]);

                if FileExists(FolderFiles[k]) then begin
                    if ZipFile.IndexOf(FileNameInZIP) > - 1 then begin
                        //if NOT Asked then begin
                            //YesCurrent := False;
                            case MessageDlg('File already exists in ZIP: ' + FileNameInZIP + #13#10 + 'Overwrite?', mtConfirmation, [mbYes, mbNo, mbCancel{, mbNoToAll, mbYesToAll}], 0) of
                                //mrYes: YesCurrent := True;
                                mrNo: Continue;
                                mrCancel: begin
                                    Result := S_OK;
                                    Exit;
                                end;
                                {
                                mrNoToAll: begin
                                    Asked := True;
                                    YesToAll := False;
                                end;
                                mrYesToAll: begin
                                    Asked := True;
                                    YesToAll := True;
                                end;
                                }
                            end;
                        //end;
                        //if YesCurrent
                        //OR YesToAll
                        //then begin
                            ZipFile.Delete(FileNameInZIP);
                        //end;
                    end;
                end;

                if ZipFile.IndexOf(ExtractFilePathZIP(FileNameInZIP)) > - 1 then begin
                    ZipFile.Delete(ExtractFilePathZIP(FileNameInZIP));
                end;

                ZipFile.Add(FolderFiles[k], FileNameInZIP);
                if Assigned(ZIPProgressCompleted) then begin
                    ZIPProgressCompleted(FolderFiles[k], NGetFileSize(FolderFiles[k]));
                end;
                Result := S_OK;
            end;
        end;
        //Result := S_OK;
    finally
        ZipFile.Close;
        FreeAndNil(ZipFile);
    end;
end;

function FOMoveFileToZIPAs(DestinationFolder: PItemIDList; Item: PItemIDList; NewFileName: String; Handle: THandle; ZipProgressEvent: TZipProgressEvent = nil; ZIPProgressCompleted: TZIPProgressCompleted = nil): HRESULT; overload;
var
    ZIPFileName: String;
begin
    Result := FOCopyFileToZIPAs(DestinationFolder, Item, NewFileName, Handle, ZipProgressEvent, ZIPProgressCompleted);
    if Result = S_OK then begin
        if IsZIPFolder(Item, ZIPFileName) then begin
            Result := FODeleteFileInZIP(Item, Handle, ZipProgressEvent, ZIPProgressCompleted);
        end else begin
            Result := FODeleteFile(Item, False, Handle);
        end;
    end;
end;

function CalculateFolderSizeFileSystem(FolderName: String): UInt64;
var
    SearchRec: TSearchRec;
begin
    Result := 0;
    try
        if FindFirst(IncludeTrailingPathDelimiter(FolderName) + '*.*', faAnyFile, SearchRec) = 0 then begin
            repeat
                if (SearchRec.Name <> '.')
                AND (SearchRec.Name <> '..')
                then begin
                    if (SearchRec.FindData.dwFileAttributes AND FILE_ATTRIBUTE_DIRECTORY) = 0 then begin
                        if SearchRec.Size > 0 then begin
                            Result := Result + UInt64(SearchRec.Size);
                        end;
                    end else begin
                        Result := Result + CalculateFolderSizeFileSystem(IncludeTrailingPathDelimiter(FolderName) + SearchRec.Name);
                    end;
                end;
            until FindNext(SearchRec) <> 0;
        end;
    finally
        System.SysUtils.FindClose(SearchRec);
    end;
end;

function FODeleteFilesInZIP(ItemIdList: TArrayOfPItemIDList; Handle: THandle; ZipProgressEvent: TZipProgressEvent = nil; ZIPProgressCompleted: TZIPProgressCompleted = nil): HRESULT; overload;
var
    IsZIPFolderDestination: Boolean;
    ZIPFileName: String;
    ZipFile: TZipFile;
    i, k: Integer;
    FileNamesInZIP: TStringList;
    FileNameInZIP: String;
    Header: TZipHeader;
    CurrentZIPFileName: String;
begin
    Result := S_FALSE;
    if (NOT Assigned(ItemIdList))
    OR (Length(ItemIdList) = 0)
    then begin
        Exit;
    end;
    IsZIPFolderDestination := IsZIPFolder(ItemIdList[0], ZIPFileName);
    if NOT IsZIPFolderDestination then begin
        Exit;
    end;
    FileNamesInZIP := TStringList.Create;
    ZipFile := TZipFile.Create;
    try
        //* Important to parse the PIDLs before the actual ZIP file is opened!
        for i := 0 to High(ItemIdList) do begin
            FileNamesInZIP.Append(ZIPFolderPath(ItemIdList[i]));
        end;
        ZipFile.OnProgress := ZipProgressEvent;
        ZipFile.Open(ZIPFileName, zmReadWrite);
        for i := 0 to FileNamesInZIP.Count - 1 do begin
            FileNameInZIP := FileNamesInZIP[i];
            for k := ZipFile.FileCount - 1 downto 0 do begin
                //* "/" check
                if SameText(ZipFile.FileName[k], FileNameInZIP)
                OR ZipFile.FileName[k].StartsWith(FileNameInZIP + '/', True)
                //* "\" check, not sure about this, ZIP files created on UNIX systems may support this???
                //OR SameText(ZipFile.FileName[k], StringReplace(FileNameInZIP, '/', '\', [rfReplaceAll]))
                //OR ZipFile.FileName[k].StartsWith(StringReplace(FileNameInZIP, '/', '\', [rfReplaceAll]) + '\', True)
                then begin
                    CurrentZIPFileName := ZipFile.FileName[k];
                    Header := ZipFile.FileInfo[k];
                    if Assigned(ZipProgressEvent) then begin
                        ZipProgressEvent(ZipFile, CurrentZIPFileName, Header, 0);
                    end;
                    ZipFile.Delete(k);
                    if Assigned(ZIPProgressCompleted) then begin
                        ZIPProgressCompleted(CurrentZIPFileName, Header.UncompressedSize);
                    end;
                end;
            end;
        end;
        ZipFile.Close;
        Result := S_OK;
    finally
        FreeAndNil(ZipFile);
        FreeAndNil(FileNamesInZIP);
    end;
end;

function FODeleteFileInZIP(Item: PItemIDList; Handle: THandle; ZipProgressEvent: TZipProgressEvent = nil; ZIPProgressCompleted: TZIPProgressCompleted = nil): HRESULT; overload;
var
    IsZIPFolderDestination: Boolean;
    ZIPFileName: String;
    ZipFile: TZipFile;
    k: Integer;
    FileNameInZIP: String;
    Header: TZipHeader;
    CurrentZIPFileName: String;
begin
    Result := S_FALSE;
    if NOT Assigned(Item) then begin
        Exit;
    end;
    IsZIPFolderDestination := IsZIPFolder(Item, ZIPFileName);
    if NOT IsZIPFolderDestination then begin
        Exit;
    end;
    if Assigned(Item) then begin
        ZipFile := TZipFile.Create;
        try
            //* Important to parse the PIDLs before the actual ZIP file is opened!
            FileNameInZIP := ZIPFolderPath(Item);
            ZipFile.OnProgress := ZipProgressEvent;
            ZipFile.Open(ZIPFileName, zmReadWrite);
            for k := ZipFile.FileCount - 1 downto 0 do begin
                //* "/" check
                if SameText(ZipFile.FileName[k], FileNameInZIP)
                OR ZipFile.FileName[k].StartsWith(FileNameInZIP + '/', True)
                //* "\" check, not sure about this, ZIP files created on UNIX systems may support this???
                //OR SameText(ZipFile.FileName[k], StringReplace(FileNameInZIP, '/', '\', [rfReplaceAll]))
                //OR ZipFile.FileName[k].StartsWith(StringReplace(FileNameInZIP, '/', '\', [rfReplaceAll]) + '\', True)
                then begin
                    CurrentZIPFileName := ZipFile.FileName[k];
                    Header := ZipFile.FileInfo[k];
                    if Assigned(ZipProgressEvent) then begin
                        ZipProgressEvent(ZipFile, CurrentZIPFileName, Header, 0);
                    end;
                    ZipFile.Delete(k);
                    if Assigned(ZIPProgressCompleted) then begin
                        ZIPProgressCompleted(CurrentZIPFileName, Header.UncompressedSize);
                    end;
                end;
            end;
            ZipFile.Close;
            Result := S_OK;
        finally
            FreeAndNil(ZipFile);
        end;
    end;
end;

function GetPIDLFileSize(Item: PItemIDList): UInt64;
var
    Error: HRESULT;
    SiItem: IShellItem;
    SiItem2: IShellItem2;
begin
    Result := 0;
    Error := SHCreateItemFromIDList(Item, IShellItem, SiItem);
    if Succeeded(Error) then begin
        SiItem2 := SiItem as IShellItem2;
        SiItem2.GetUInt64(PKEY_Size, Result);
    end;
end;

function CalculateFolderSizePIDL(PIDL: PItemIDList; Handle: HWND): UInt64;
var
    Desktop: IShellFolder;
    EnumIDLIst: IEnumIDList;
    Fetched: ULONG;
    CurPIDL: PItemIDList;
    NewPIDL: PItemIDList;
    Attributes: Cardinal;
    Error: Integer;
    Flags: Cardinal;
    CurrentShellFolder: IShellFolder;
    FolderName: String;
    ItemName: String;
    PIDLFileSize: UInt64;
    FileSize: Int64;
begin
    Result := 0;
    if SHGetDesktopFolder(Desktop) = NOERROR then begin
        Flags := SHCONTF_FOLDERS OR SHCONTF_NONFOLDERS OR SHCONTF_INCLUDEHIDDEN;
        Error := Desktop.BindToObject(PIDL, nil, IID_IShellFolder, Pointer(CurrentShellFolder));
        if Error = NOERROR then begin
            CurrentShellFolder.EnumObjects(Handle, Flags, EnumIDList);
            if Assigned(EnumIDList) then begin
                while EnumIDList.Next(1, CurPIDL, Fetched) = S_OK do begin
                    try
                        Attributes := SFGAO_FOLDER OR SFGAO_FILESYSANCESTOR OR SFGAO_FILESYSTEM {OR SFGAO_COMPRESSED};
                        CurrentShellFolder.GetAttributesOf(1, CurPIDL, Attributes);
                        NewPIDL := ILCombine(PIDL, CurPIDL);
                        try
                            if ((Attributes AND SFGAO_FOLDER) > 0) then begin
                                FolderName := GetFileNameFromPIDL(NewPIDL);
                                if DirectoryExists(FolderName) then begin
                                    Result := Result + CalculateFolderSizeFileSystem(FolderName);
                                end else begin
                                    if FileExists(FolderName) then begin
                                        PIDLFileSize := NGetFileSize(ItemName);
                                    end else begin
                                        PIDLFileSize := GetPIDLFileSize(NewPIDL);
                                    end;
                                    //* Not archive file possibly if file size is 0
                                    if PIDLFileSize = 0 then begin
                                        Result := Result + CalculateFolderSizePIDL(NewPIDL, Handle);
                                    end else begin
                                        Result := Result + PIDLFileSize;
                                    end;
                                end;
                            end else begin
                                ItemName := GetFileNameFromPIDL(NewPIDL);
                                if FileExists(ItemName) then begin
                                    FileSize := NGetFileSize(ItemName);
                                    if FileSize > 0 then begin
                                        Result := Result + UInt64(FileSize);
                                    end;
                                end else begin
                                    Result := Result + GetPIDLFileSize(NewPIDL);
                                end;
                            end;
                        finally
                            CoTaskMemFree(NewPIDL);
                        end;
                    finally
                        CoTaskMemFree(CurPIDL);
                    end;
                end;
            end;
        end;
    end;
end;

function CalculateFolderSizeItemIDList(ItemIdList: TArrayOfPItemIDList; Handle: HWND): UInt64;
var
    i: Integer;
    FileName: String;
    PIDLFileSize: UInt64;
begin
    Result := 0;
    for i := 0 to High(ItemIdList) do begin
        FileName := GetFileNameFromPIDL(ItemIdList[i]);
        if FileExists(FileName) then begin
            Result := Result + NGetFileSize(FileName);
        end else if DirectoryExists(FileName) then begin
            Result := Result + CalculateFolderSizeFileSystem(FileName);
        end else begin
            PIDLFileSize := GetPIDLFileSize(ItemIdList[i]);
            if PIDLFileSize <> 0 then begin
                Result := Result + PIDLFileSize
            end else begin
                Result := Result + CalculateFolderSizePIDL(ItemIdList[i], Handle);
            end;
        end;
    end;
end;

function CalculateFolderSizeFileSystemList(FileList: TFOFileList; Handle: HWND): UInt64;
var
    i: Integer;
    FileName: String;
begin
    Result := 0;
    for i := Low(FileList) to High(FileList) do begin
        FileName := FileList[i];
        if FileExists(FileName) then begin
            Result := Result + NGetFileSize(FileName);
        end else if DirectoryExists(FileName) then begin
            Result := Result + CalculateFolderSizeFileSystem(FileName);
        end;
    end;
end;

{$POINTERMATH ON}

function HIDA_GetPIDLFolder(pida: PIDA): LPITEMIDLIST;
begin
    Result := LPITEMIDLIST(LPBYTE(pida) + pida.aoffset[0]);
end;

function HIDA_GetPIDLItem(pida: PIDA; idx: Integer): LPITEMIDLIST;
begin
    Result := LPITEMIDLIST(LPBYTE(pida) + (PUINT(@pida.aoffset[0]) + (1 + idx))^);
end;

type
    CidaPidlInfo = record
        Pidl: PItemIDList;
        PidlOffset: UINT;
        PidlSize: UINT;
    end;

procedure CopyFileNamesToClipboard(FileNames: TFOFileList; CutMode: Boolean);
var
    PidlInfo: array of CidaPidlInfo;
    Attrs: Cardinal;
    AllocSize: NativeUInt;
    gmem: THandle;
    ida: PIDA;
    i: Integer;
    PCut: PCardinal;
begin
    if (Length(FileNames) = 0)
    OR (CF_SHELLIDLIST = 0)
    then begin
        Exit;
    end;
    SetLength(PidlInfo, Length(FileNames));
    for i := Low(PidlInfo) to High(PidlInfo) do begin
        PidlInfo[i].Pidl := nil;
    end;
    try
        AllocSize := SizeOf(CIDA) + (SizeOf(UINT) * Length(FileNames)) + SizeOf(Word);
        for i := 0 to Length(FileNames) - 1 do begin
            OleCheck(SHParseDisplayName(PChar(FileNames[i]), nil, PidlInfo[i].Pidl, 0, Attrs));
            PidlInfo[i].PidlOffset := AllocSize;
            PidlInfo[i].PidlSize := ILGetSize(PidlInfo[i].Pidl);
            Inc(AllocSize, PidlInfo[i].PidlSize);
        end;
        gmem := GlobalAlloc(GMEM_MOVEABLE, AllocSize);
        if gmem = 0 then RaiseLastOSError;
        try
            ida := PIDA(GlobalLock(gmem));
            if ida = nil then begin
                RaiseLastOSError;
            end;
            ida.cidl := Length(FileNames);
            ida.aoffset[0] := SizeOf(CIDA) + (SizeOf(UINT) * Length(FileNames));
            HIDA_GetPIDLFolder(ida).mkid.cb := 0;
            for i := 0 to Length(FileNames) - 1 do begin
                ida.aoffset[1 + i] := PidlInfo[i].PidlOffset;
                Move(PidlInfo[i].Pidl^, HIDA_GetPIDLItem(ida, i)^, PidlInfo[i].PidlSize);
            end;
            GlobalUnlock(gmem);
            Clipboard.SetAsHandle(CF_SHELLIDLIST, gmem);
        except
            GlobalFree(gmem);
            raise;
        end;
        try
            if CutMode then begin
                gmem := GlobalAlloc(GMEM_MOVEABLE, SizeOf(MOVEEFFECTBYTESVALUE));
                if gmem = 0 then begin
                    RaiseLastOSError;
                end;
                PCut := GlobalLock(gmem);
                PCut^ := MOVEEFFECTBYTESVALUE;
                GlobalUnlock(gmem);
                Clipboard.SetAsHandle(CF_PREFERREDDROPEFFECT, gmem);
            end;
        except
            GlobalFree(gmem);
            raise;
        end;
    finally
        for i := Low(PidlInfo) to High(PidlInfo) do begin
            CoTaskMemFree(PidlInfo[i].Pidl);
        end;
    end;
end;

procedure CopyPIDLsToClipboard(ItemIdList: TArrayOfPItemIDList; CutMode: Boolean);
var
    PidlInfo: array of CidaPidlInfo;
    AllocSize: NativeUInt;
    gmem: THandle;
    ida: PIDA;
    i: Integer;
    PCut: PCardinal;
begin
    if (ItemIdList = nil)
    OR (Length(ItemIdList) = 0)
    OR (CF_SHELLIDLIST = 0)
    then begin
        Exit;
    end;
    SetLength(PidlInfo, Length(ItemIdList));
    for i := Low(PidlInfo) to High(PidlInfo) do begin
        PidlInfo[i].Pidl := nil;
    end;
    AllocSize := SizeOf(CIDA) + (SizeOf(UINT) * Length(ItemIdList)) + SizeOf(Word);
    for i := 0 to High(ItemIdList) do begin
        PidlInfo[i].Pidl := ItemIdList[i];
        PidlInfo[i].PidlOffset := AllocSize;
        PidlInfo[i].PidlSize := ILGetSize(PidlInfo[i].Pidl);
        Inc(AllocSize, PidlInfo[i].PidlSize);
    end;
    Clipboard.Open;
    try
        gmem := GlobalAlloc(GMEM_MOVEABLE, AllocSize);
        if gmem = 0 then RaiseLastOSError;
        try
            ida := PIDA(GlobalLock(gmem));
            if ida = nil then begin
                RaiseLastOSError;
            end;
            ida.cidl := Length(ItemIdList);
            ida.aoffset[0] := SizeOf(CIDA) + (SizeOf(UINT) * Length(ItemIdList));
            HIDA_GetPIDLFolder(ida).mkid.cb := 0;
            for i := 0 to High(ItemIdList) do begin
                ida.aoffset[1 + i] := PidlInfo[i].PidlOffset;
                Move(PidlInfo[i].Pidl^, HIDA_GetPIDLItem(ida, i)^, PidlInfo[i].PidlSize);
            end;
            GlobalUnlock(gmem);
            Clipboard.SetAsHandle(CF_SHELLIDLIST, gmem);
        except
            GlobalFree(gmem);
            raise;
        end;
        try
            if CutMode then begin
                gmem := GlobalAlloc(GMEM_MOVEABLE, SizeOf(MOVEEFFECTBYTESVALUE));
                if gmem = 0 then begin
                    RaiseLastOSError;
                end;
                PCut := GlobalLock(gmem);
                PCut^ := MOVEEFFECTBYTESVALUE;
                GlobalUnlock(gmem);
                Clipboard.SetAsHandle(CF_PREFERREDDROPEFFECT, gmem);
            end;
        except
            GlobalFree(gmem);
            raise;
        end;
    finally
        Clipboard.Close;
    end;
end;

function GetPIDLsFromData(Data: pointer; Size: integer; var ItemIdList: TArrayOfPItemIDList): Boolean;
var
    i: integer;
    pOffset: ^UINT;
    PIDL: PItemIDList;
    PIDLCounter: Integer;
    DesktopPIDL: PItemIDList;
    RootPIDL: PItemIDList;
begin
    SetLength(ItemIdList, 0);
    Result := (Data <> nil) AND (Size >= integer(PIDA(Data)^.cidl) * (SizeOf(UINT) + SizeOf(PItemIDList)) + SizeOf(UINT));
    if NOT Result then begin
        Exit;
    end;
    RootPIDL := nil;
    SHGetKnownFolderIDList(FOLDERID_Desktop, 0, 0, DesktopPIDL);
    try
        pOffset := @(PIDA(Data)^.aoffset[0]);
        i := PIDA(Data)^.cidl; // Note: Count doesn't include folder PIDL
        SetLength(ItemIdList, i);
        PIDLCounter := 0;
        while (i >= 0) do begin
            PIDL := PItemIDList(Data);
            inc(PByte(PIDL), pOffset^);
            //Showmessage(GetFileNameFromPIDL(PIDL));
            if PIDLCounter = 0 then begin
                if NOT ILIsEqual(PIDL, DesktopPIDL) then begin
                    RootPIDL := PIDL;
                end;
            end;
            if PIDLCounter > 0 then begin
                if Assigned(RootPIDL) then begin
                    ItemIdList[PIDLCounter - 1] := ILCombine(RootPIDL, PIDL);
                end else begin
                    ItemIdList[PIDLCounter - 1] := ILClone(PIDL);
                end;
            end;
            Inc(pOffset);
            Dec(i);
            Inc(PIDLCounter);
        end;
    finally
        CoTaskMemFree(DesktopPIDL);
    end;
    Result := PIDLCounter > 1;
end;

function GetPIDLsFromHGlobal(const HGlob: HGlobal; var ItemIdList: TArrayOfPItemIDList): Boolean;
var
    pCIDA: PIDA;
begin
    pCIDA := PIDA(GlobalLock(HGlob));
    try
        Result := GetPIDLsFromData(pCIDA, GlobalSize(HGlob), ItemIdList);
    finally
        GlobalUnlock(HGlob);
    end;
end;

function GetPIDLsFromClipboard(var ItemIdList: TArrayOfPItemIDList; var CutMode: Boolean): Boolean;
var
    DataObject: IDataObject;
    HR: HResult;
    Format: TFormatEtc;
    FileContentsMedium: TStgMedium;
    PCut: PByte;
    CutContentsMedium: TStgMedium;
begin
    Result := False;
    CutMode := False;
    OleGetClipboard(DataObject);
    //* Get PIDL list
    ZeroMemory(@Format, SizeOf(TFormatEtc));
    Format.cfFormat := CF_SHELLIDLIST;
    Format.dwAspect := DVASPECT_CONTENT;
    Format.lindex := - 1;
    Format.tymed := TYMED_HGLOBAL;
    ZeroMemory(@FileContentsMedium, SizeOf(FileContentsMedium));
    HR := DataObject.GetData(Format, FileContentsMedium);
    if Failed(HR) then begin
        Exit;
    end;
    try
        Result := GetPIDLsFromHGlobal(FileContentsMedium.hGlobal, ItemIdList);
    finally
        ReleaseStgMedium(FileContentsMedium);
    end;
    //* Cut?
    ZeroMemory(@Format, SizeOf(TFormatEtc));
    Format.cfFormat := CF_PREFERREDDROPEFFECT;
    Format.dwAspect := DVASPECT_CONTENT;
    Format.lindex := - 1;
    Format.tymed := TYMED_HGLOBAL;
    ZeroMemory(@CutContentsMedium, SizeOf(CutContentsMedium));
    HR := DataObject.GetData(Format, CutContentsMedium);
    if Failed(HR) then begin
        Exit;
    end;
    PCut := GlobalLock(CutContentsMedium.hGlobal);
    try
        if PCut^ = MOVEEFFECTBYTESVALUE then begin
            CutMode := True;
        end;
    finally
        GlobalUnlock(CutContentsMedium.hGlobal);
    end;
end;

initialization

    CF_SHELLIDLIST := RegisterClipboardFormat(CFSTR_SHELLIDLIST);
    CF_PREFERREDDROPEFFECT := RegisterClipboardFormat(CFSTR_PREFERREDDROPEFFECT);

end.
