//********************************************************************************************************************************
//*                                                                                                                              *
//*     File Operations Library 1.0.10.10  3delite 2023-2024                                                                    *
//*     See File Operations Library ReadMe.txt for details                                                                       *
//*                                                                                                                              *
//* Licenses available if you like and use this library:                                                                         *
//* Freeware License: 15                                                                                                        *
//*     http://www.shareit.com/product.html?productid=301063451                                                                  *
//* Shareware License: 25                                                                                                       *
//*     http://www.shareit.com/product.html?productid=301063452                                                                  *
//* Commercial License: 95                                                                                                      *
//*     http://www.shareit.com/product.html?productid=301063450                                                                  *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object%20Pascal%20Developer%20Resources/FileOperationsLibrary.html                                *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************

unit FileOperationsSearchLibrary;

{$WARN SYMBOL_PLATFORM OFF}

interface

Uses
    SysUtils,
    Classes,
    Windows,
    SHLObj,
    MultiThread2;

type
    TFOSearchMatchCallback = procedure(FileName: String) of object;
    TFOSearchProgressCallback = procedure(FileName: String) of object;
    TFOSearchFinishedCallback = procedure of object;

type
    TFOSearchValueCharacterEncoding = (svceANSIUTF8Unicode, svceANSI, svceUTF8, svceUnicode);

type
    TFOSearchThread = class;

    TFOSearchOptionsInternalData = record
        ANSIBytes: TBytes;
        UnicodeBytes: TBytes;
        UnicodeUpperCaseValue: String;
        UTF8Bytes: TBytes;
        MaxBytes: Integer;
        LastReportTime: TDateTime;
        SearchThread: TFOSearchThread;
    end;

    TFOSearchOptions = class
        FileName: String;
        FileNameMatchCase: Boolean;
        MatchFolderNames: Boolean;
        Value: String;
        ValueMatchCase: Boolean;
        ValueCharacterEncoding: TFOSearchValueCharacterEncoding;
        SubFolders: Boolean;
        FileSizeRange: Boolean;
        FileSizeMin: UInt64;
        FileSizeMax: UInt64;
        DateTimeRange: Boolean;
        DateTimeMin: TDateTime;
        DateTimeMax: TDateTime;
        MultiThread: Boolean;
        SearchMatchCallback: TFOSearchMatchCallback;
        SearchProgressCallback: TFOSearchProgressCallback;
        SearchFinishedCallback: TFOSearchFinishedCallback;
        InternalData: TFOSearchOptionsInternalData;
    end;

    TSearchScanJob = class
        FileName: String;
    end;

    TFOSearchThread = class(TThread)
    private
        FPIDL: PItemIDList;
        FSearchOptions: TFOSearchOptions;
        FSearchThreadCount: Integer;
        FHandle: HWND;
        FMultiThread2: TMultiThread2;
        procedure ThreadWorkerCallback(Thread: TMultiThreadThread; var Parameters: TObject; var Job: TObject; var Results: TObject);
        procedure CheckToStartScanThreads;
    protected
        procedure Execute; override;
    public
        Constructor Create(PIDL: PItemIDList; SearchOptions: TFOSearchOptions; SearchThreadCount: Integer; Handle: HWND); overload;
        Destructor Destroy; override;
    end;

    function SearchFolderFileSystem(FolderName: String; SearchOptions: TFOSearchOptions; Thread: TThread = nil): Boolean;
    function SearchFolderPIDL(PIDL: PItemIDList; SearchOptions: TFOSearchOptions; Handle: HWND; Thread: TThread = nil): Boolean;

implementation

Uses
    ActiveX,
    IOUtils,
    Math,
    DateUtils,
    System.Character,
    FileOperationsLibrary;

function SearchCheckFileContents(FileName: String; SearchOptions: TFOSearchOptions; Thread: TThread = nil): Boolean;
Const
    BUFFER_SIZE = 1024 * 1024 * 1024;
var
    FileStream: TFileStream;
    Buffer: PByte;
    BufferSize: Integer;
    SearchBytes: TBytes;
    i, k: Integer;
    PBuffer: PByte;
    PBufferChar: PByte;
    PBufferChar2: PWord;
    MatchFound: Boolean;
    SearchString: String;
    PBufferSearchFor: PByte;
    UTF8Character: Char;
    FirstByte, SecondByte: Byte;
    PBufferNext: PByte;
begin
    Result := False;
    //* Create search bytes if not yet created
    if (SearchOptions.ValueCharacterEncoding = svceANSIUTF8Unicode)
    OR (SearchOptions.ValueCharacterEncoding = svceANSI)
    then begin
        if Length(SearchOptions.InternalData.ANSIBytes) = 0 then begin
            if SearchOptions.ValueMatchCase then begin
                SearchOptions.InternalData.ANSIBytes := TEncoding.ANSI.GetBytes(SearchOptions.Value);
            end else begin
                SearchOptions.InternalData.ANSIBytes := TEncoding.ANSI.GetBytes(WideUpperCase(SearchOptions.Value));
            end;
        end;
    end;
    if (SearchOptions.ValueCharacterEncoding = svceANSIUTF8Unicode)
    OR (SearchOptions.ValueCharacterEncoding = svceUTF8)
    then begin
        if Length(SearchOptions.InternalData.UTF8Bytes) = 0 then begin
            if SearchOptions.ValueMatchCase then begin
                SearchOptions.InternalData.UTF8Bytes := TEncoding.UTF8.GetBytes(SearchOptions.Value);
            end else begin
                SearchOptions.InternalData.UTF8Bytes := TEncoding.UTF8.GetBytes(WideUpperCase(SearchOptions.Value));
                SearchOptions.InternalData.UnicodeUpperCaseValue := WideUpperCase(SearchOptions.Value);
            end;
        end;
    end;
    if (SearchOptions.ValueCharacterEncoding = svceANSIUTF8Unicode)
    OR (SearchOptions.ValueCharacterEncoding = svceUnicode)
    then begin
        if Length(SearchOptions.InternalData.UnicodeBytes) = 0 then begin
            if SearchOptions.ValueMatchCase then begin
                SearchOptions.InternalData.UnicodeBytes := TEncoding.Unicode.GetBytes(SearchOptions.Value);
            end else begin
                SearchOptions.InternalData.UnicodeBytes := TEncoding.Unicode.GetBytes(WideUpperCase(SearchOptions.Value));
                SearchOptions.InternalData.UnicodeUpperCaseValue := WideUpperCase(SearchOptions.Value);
            end;
        end;
    end;
    if SearchOptions.InternalData.MaxBytes = 0 then begin
        if SearchOptions.ValueCharacterEncoding = svceANSIUTF8Unicode then begin
            SearchOptions.InternalData.MaxBytes := Length(SearchOptions.InternalData.UnicodeBytes);
        end;
        if SearchOptions.ValueCharacterEncoding = svceANSI then begin
            SearchOptions.InternalData.MaxBytes := Length(SearchOptions.InternalData.ANSIBytes)
        end;
        if SearchOptions.ValueCharacterEncoding = svceUTF8 then begin
            SearchOptions.InternalData.MaxBytes := Length(SearchOptions.InternalData.UTF8Bytes);
        end;
        if SearchOptions.ValueCharacterEncoding = svceUnicode then begin
            SearchOptions.InternalData.MaxBytes := Length(SearchOptions.InternalData.UnicodeBytes);
        end;
    end;
    //* Scan the file
    try
        FileStream := TFileStream.Create(FileName, fmOpenRead);
        try
            GetMem(Buffer, BUFFER_SIZE + 1); //* +1 for UTF-8
            try
                MatchFound := False;
                if FileStream.Size - FileStream.Position < BUFFER_SIZE then begin
                    BufferSize := FileStream.Read(Buffer^, FileStream.Size - FileStream.Position);
                end else begin
                    BufferSize := FileStream.Read(Buffer^, BUFFER_SIZE);
                end;
                //* Zero the last +1 buffer value for UTF-8
                PBuffer := Buffer;
                Inc(PBuffer, BufferSize);
                PBuffer^ := 0;
                if (SearchOptions.ValueCharacterEncoding = svceANSIUTF8Unicode)
                OR (SearchOptions.ValueCharacterEncoding = svceANSI)
                then begin
                    SearchBytes := SearchOptions.InternalData.ANSIBytes;
                    PBuffer := Buffer;
                    if SearchOptions.ValueMatchCase then begin
                        for i := 0 to BufferSize - Length(SearchBytes) {- 1} do begin
                            PBufferChar := PBuffer;
                            for k := 0 to High(SearchBytes) do begin
                                if PBufferChar^ <> SearchBytes[k] then begin
                                    Break;
                                end;
                                if k = High(SearchBytes) then begin
                                    MatchFound := True;
                                end;
                                Inc(PBufferChar);
                            end;
                            if MatchFound then begin
                                Break;
                            end;
                            Inc(PBuffer);
                        end;
                    end else begin
                        for i := 0 to BufferSize - Length(SearchBytes) {- 1} do begin
                            PBufferChar := PBuffer;
                            for k := 0 to High(SearchBytes) do begin
                                if Ord(UpCase(ANSIChar(PBufferChar^))) <> SearchBytes[k] then begin
                                    Break;
                                end;
                                if k = High(SearchBytes) then begin
                                    MatchFound := True;
                                end;
                                Inc(PBufferChar);
                            end;
                            if MatchFound then begin
                                Break;
                            end;
                            Inc(PBuffer);
                        end;
                    end;
                end;
                if Assigned(Thread) then begin
                    if Thread.CheckTerminated then begin
                        Exit;
                    end;
                end;
                if MatchFound then begin
                    SearchOptions.SearchMatchCallback(FileName);
                    Exit;
                end;
                if (SearchOptions.ValueCharacterEncoding = svceANSIUTF8Unicode)
                OR (SearchOptions.ValueCharacterEncoding = svceUnicode)
                then begin
                    SearchBytes := SearchOptions.InternalData.UnicodeBytes;
                    PBuffer := Buffer;
                    if SearchOptions.ValueMatchCase then begin
                        for i := 0 to BufferSize - Length(SearchBytes) {- 1} do begin
                            PBufferChar := PBuffer;
                            for k := 0 to High(SearchBytes) do begin
                                if PBufferChar^ <> SearchBytes[k] then begin
                                    Break;
                                end;
                                if k = High(SearchBytes) then begin
                                    MatchFound := True;
                                end;
                                Inc(PBufferChar);
                            end;
                            if MatchFound then begin
                                Break;
                            end;
                            Inc(PBuffer);
                        end;
                    end else begin
                        SearchString := SearchOptions.InternalData.UnicodeUpperCaseValue;
                        for i := 0 to BufferSize - (Length(SearchString) * 2) {- 1} do begin
                            PBufferChar2 := PWord(PBuffer);
                            for k := Low(SearchString) to High(SearchString) do begin
                                if Char(PBufferChar2^).ToUpper <> SearchString[k] then begin
                                    Break;
                                end;
                                if k = High(SearchString) then begin
                                    MatchFound := True;
                                end;
                                Inc(PBufferChar2);
                            end;
                            if MatchFound then begin
                                Break;
                            end;
                            Inc(PBuffer);
                        end;
                    end;
                end;
                if Assigned(Thread) then begin
                    if Thread.CheckTerminated then begin
                        Exit;
                    end;
                end;
                if MatchFound then begin
                    SearchOptions.SearchMatchCallback(FileName);
                    Exit;
                end;
                if (SearchOptions.ValueCharacterEncoding = svceANSIUTF8Unicode)
                OR (SearchOptions.ValueCharacterEncoding = svceUTF8)
                then begin
                    SearchBytes := SearchOptions.InternalData.UTF8Bytes;
                    PBuffer := Buffer;
                    if SearchOptions.ValueMatchCase then begin
                        for i := 0 to BufferSize - Length(SearchBytes) {- 1} do begin
                            PBufferChar := PBuffer;
                            for k := 0 to High(SearchBytes) do begin
                                if PBufferChar^ <> SearchBytes[k] then begin
                                    Break;
                                end;
                                if k = High(SearchBytes) then begin
                                    MatchFound := True;
                                end;
                                Inc(PBufferChar);
                            end;
                            if MatchFound then begin
                                Break;
                            end;
                            Inc(PBuffer);
                        end;
                    end else begin
                        SearchString := SearchOptions.InternalData.UnicodeUpperCaseValue;
                        for i := 0 to BufferSize - Length(SearchOptions.InternalData.UTF8Bytes) {- 1} do begin
                            PBufferSearchFor := PBuffer;
                            for k := Low(SearchString) to High(SearchString) do begin
                                FirstByte := PBufferSearchFor^;
                                if (FirstByte <= $7F) or (($C2 <= FirstByte) and (FirstByte <= $FD)) then begin
                                    PBufferNext := PBufferSearchFor;
                                    Inc(PBufferNext);
                                    SecondByte := PBufferNext^;
                                    if ($80 <= SecondByte) and (SecondByte <= $BF) then begin
                                        try
                                            UTF8Character := TEncoding.UTF8.GetString(TBytes(PBufferSearchFor), 0, 2)[1];
                                        except
                                            //*
                                        end;
                                        Inc(PBufferSearchFor, 2);
                                    end else begin
                                        UTF8Character := Char(FirstByte);
                                        Inc(PBufferSearchFor);
                                    end;
                                end else begin
                                    UTF8Character := Char(FirstByte);
                                    Inc(PBufferSearchFor);
                                end;
                                if UTF8Character.IsDefined then begin
                                    if UTF8Character.ToUpper <> SearchString[k] then begin
                                        Break;
                                    end;
                                end;
                                if k = High(SearchString) then begin
                                    MatchFound := True;
                                end;
                            end;
                            if MatchFound then begin
                                Break;
                            end;
                            Inc(PBuffer);
                        end;
                    end;
                end;
                if Assigned(Thread) then begin
                    if Thread.CheckTerminated then begin
                        Exit;
                    end;
                end;
                if MatchFound then begin
                    SearchOptions.SearchMatchCallback(FileName);
                    Exit;
                end;
                FileStream.Seek(- (SearchOptions.InternalData.MaxBytes + 2), soCurrent);
            finally
                FreeMem(Buffer);
            end;
        finally
            FreeAndNil(FileStream);
        end;
    except
        //*
    end;
end;

function SearchCheckFileContentsMultiThreaded(FileName: String; SearchOptions: TFOSearchOptions; Thread: TThread = nil): Boolean;
var
    SearchScanJob: TSearchScanJob;
begin
    SearchOptions.InternalData.SearchThread.CheckToStartScanThreads;
    SearchScanJob := TSearchScanJob.Create;
    SearchScanJob.FileName := FileName;
    SearchOptions.InternalData.SearchThread.FMultiThread2.AddWork(SearchScanJob);
    Result := True;
end;

function SearchFolderCheckFile(FileName: String; SearchOptions: TFOSearchOptions; Thread: TThread = nil): Boolean;
var
    SearchRec: TSearchRec;
begin
    Result := False;
    if FindFirst(FileName, faAnyFile, SearchRec) = 0 then begin
        if TPath.MatchesPattern(SearchRec.Name, SearchOptions.FileName, SearchOptions.FileNameMatchCase) then begin
            if SearchOptions.DateTimeRange then begin
                if DateTimeInRange(SearchRec.TimeStamp, SearchOptions.DateTimeMin, SearchOptions.DateTimeMax) then begin
                    if SearchOptions.FileSizeRange then begin
                        if InRange(UInt64(SearchRec.Size), SearchOptions.FileSizeMin, SearchOptions.FileSizeMax) then begin
                            if SearchOptions.Value = '' then begin
                                SearchOptions.SearchMatchCallback(FileName);
                            end else begin
                                if SearchOptions.MultiThread then begin
                                    SearchCheckFileContentsMultiThreaded(FileName, SearchOptions, Thread);
                                end else begin
                                    SearchCheckFileContents(FileName, SearchOptions);
                                end;
                            end;
                        end;
                    end else begin
                        if SearchOptions.Value = '' then begin
                            SearchOptions.SearchMatchCallback(FileName);
                        end else begin
                            if SearchOptions.MultiThread then begin
                                SearchCheckFileContentsMultiThreaded(FileName, SearchOptions, Thread);
                            end else begin
                                SearchCheckFileContents(FileName, SearchOptions);
                            end;
                        end;
                    end;
                    Result := True;
                end;
            end else begin
                if SearchOptions.FileSizeRange then begin
                    if InRange(UInt64(SearchRec.Size), SearchOptions.FileSizeMin, SearchOptions.FileSizeMax) then begin
                        if SearchOptions.Value = '' then begin
                            SearchOptions.SearchMatchCallback(FileName);
                        end else begin
                            if SearchOptions.MultiThread then begin
                                SearchCheckFileContentsMultiThreaded(FileName, SearchOptions, Thread);
                            end else begin
                                SearchCheckFileContents(FileName, SearchOptions);
                            end;
                        end;
                    end;
                end else begin
                    if SearchOptions.Value = '' then begin
                        SearchOptions.SearchMatchCallback(FileName);
                    end else begin
                        if SearchOptions.MultiThread then begin
                            SearchCheckFileContentsMultiThreaded(FileName, SearchOptions, Thread);
                        end else begin
                            SearchCheckFileContents(FileName, SearchOptions);
                        end;
                    end;
                    Result := True;
                end;
            end;
        end;
        System.SysUtils.FindClose(SearchRec);
    end;
end;

function SearchCheckFolder(FolderName: String; SearchOptions: TFOSearchOptions): Boolean;
var
    SearchRec: TSearchRec;
begin
    Result := False;
    if FindFirst(FolderName, faAnyFile, SearchRec) = 0 then begin
        if TPath.MatchesPattern(SearchRec.Name, SearchOptions.FileName, SearchOptions.FileNameMatchCase) then begin
            if SearchOptions.DateTimeRange then begin
                if DateTimeInRange(SearchRec.TimeStamp, SearchOptions.DateTimeMin, SearchOptions.DateTimeMax) then begin
                    if SearchOptions.FileSizeRange then begin
                        if InRange(CalculateFolderSizeFileSystem(FolderName), SearchOptions.FileSizeMin, SearchOptions.FileSizeMax) then begin
                            Result := True;
                        end;
                    end;
                end;
            end else if SearchOptions.FileSizeRange then begin
                if InRange(CalculateFolderSizeFileSystem(FolderName), SearchOptions.FileSizeMin, SearchOptions.FileSizeMax) then begin
                    Result := True;
                end;
            end else begin
                Result := True;
            end;
        end;
        System.SysUtils.FindClose(SearchRec);
    end;
end;

procedure SearchCheckToReportProgress(const FileName: String; SearchOptions: TFOSearchOptions);
begin
    if SearchOptions.InternalData.LastReportTime = 0 then begin
        SearchOptions.InternalData.LastReportTime := Now;
    end;
    if MilliSecondsBetween(Now, SearchOptions.InternalData.LastReportTime) > 500 then begin
        SearchOptions.SearchProgressCallback(FileName);
        SearchOptions.InternalData.LastReportTime := Now;
    end;
end;

function SearchFolderFileSystem(FolderName: String; SearchOptions: TFOSearchOptions; Thread: TThread = nil): Boolean;
var
    SearchRec: TSearchRec;
begin
    Result := False;
    try
        if FindFirst(IncludeTrailingPathDelimiter(FolderName) + '*.*', faAnyFile, SearchRec) = 0 then begin
            repeat
                if Assigned(Thread) then begin
                    if Thread.CheckTerminated then begin
                        Exit;
                    end;
                end;
                if (SearchRec.Name <> '.')
                AND (SearchRec.Name <> '..')
                then begin
                    if (SearchRec.FindData.dwFileAttributes AND FILE_ATTRIBUTE_DIRECTORY) = 0 then begin
                        Result := SearchFolderCheckFile(IncludeTrailingPathDelimiter(FolderName) + SearchRec.Name, SearchOptions, Thread);
                    end else begin

                        if SearchOptions.MatchFolderNames then begin
                            if SearchCheckFolder(IncludeTrailingPathDelimiter(FolderName) + SearchRec.Name, SearchOptions) then begin
                                SearchOptions.SearchMatchCallback(IncludeTrailingPathDelimiter(FolderName) + SearchRec.Name);
                            end;
                        end;

                        SearchCheckToReportProgress(IncludeTrailingPathDelimiter(FolderName) + SearchRec.Name, SearchOptions);

                        if SearchOptions.SubFolders then begin
                            Result := SearchFolderFileSystem(IncludeTrailingPathDelimiter(FolderName) + SearchRec.Name, SearchOptions);
                        end;
                    end;
                end;
            until FindNext(SearchRec) <> 0;
        end;
    finally
        System.SysUtils.FindClose(SearchRec);
    end;
end;

function SearchFolderPIDL(PIDL: PItemIDList; SearchOptions: TFOSearchOptions; Handle: HWND; Thread: TThread = nil): Boolean;
var
    Desktop: IShellFolder;
    EnumIDLIst: IEnumIDList;
    Fetched: ULONG;
    CurPIDL: PItemIDList;
    NewPIDL: PItemIDList;
    Attributes: Cardinal;
    Error: Integer;
    Flags: Cardinal;
    CurrentShellFolder: IShellFolder;
    FolderName: String;
    ItemName: String;
    PIDLFileSize: Int64;
begin
    Result := False;
    if SHGetDesktopFolder(Desktop) = NOERROR then begin
        Flags := SHCONTF_FOLDERS OR SHCONTF_NONFOLDERS OR SHCONTF_INCLUDEHIDDEN;
        Error := Desktop.BindToObject(PIDL, nil, IID_IShellFolder, Pointer(CurrentShellFolder));
        if Error = NOERROR then begin
            CurrentShellFolder.EnumObjects(Handle, Flags, EnumIDList);
            if Assigned(EnumIDList) then begin
                while EnumIDList.Next(1, CurPIDL, Fetched) = S_OK do begin
                    try
                        if Assigned(Thread) then begin
                            if Thread.CheckTerminated then begin
                                Exit;
                            end;
                        end;
                        Attributes := SFGAO_FOLDER OR SFGAO_FILESYSANCESTOR OR SFGAO_FILESYSTEM {OR SFGAO_COMPRESSED};
                        CurrentShellFolder.GetAttributesOf(1, CurPIDL, Attributes);
                        NewPIDL := ILCombine(PIDL, CurPIDL);
                        try
                            if ((Attributes AND SFGAO_FOLDER) > 0)
                            AND SearchOptions.SubFolders
                            then begin
                                FolderName := GetFileNameFromPIDL(NewPIDL);

                                if SearchOptions.MatchFolderNames then begin
                                    if SearchCheckFolder(FolderName, SearchOptions) then begin
                                        SearchOptions.SearchMatchCallback(FolderName);
                                    end;
                                end;

                                SearchCheckToReportProgress(FolderName, SearchOptions);

                                if DirectoryExists(FolderName) then begin
                                    Result := SearchFolderFileSystem(FolderName, SearchOptions);
                                end else begin
                                    PIDLFileSize := 0;
                                    if FileExists(FolderName) then begin
                                        PIDLFileSize := NGetFileSize(ItemName);
                                    end;
                                    //* Not archive file possibly if file size is 0
                                    if PIDLFileSize = 0 then begin
                                        Result := SearchFolderPIDL(NewPIDL, SearchOptions, Handle);
                                    end else begin
                                        Result := SearchFolderCheckFile(FolderName, SearchOptions, Thread);
                                    end;
                                end;
                            end else begin
                                ItemName := GetFileNameFromPIDL(NewPIDL);
                                if FileExists(ItemName) then begin
                                    Result := SearchFolderCheckFile(ItemName, SearchOptions, Thread);
                                end;
                            end;
                        finally
                            CoTaskMemFree(NewPIDL);
                        end;
                    finally
                        CoTaskMemFree(CurPIDL);
                    end;
                end;
            end;
        end;
    end;
end;

{ TFOSearchThread }

procedure TFOSearchThread.CheckToStartScanThreads;
begin
    if NOT Assigned(FMultiThread2) then begin
        FMultiThread2 := TMultiThread2.Create(nil);
        FMultiThread2.ThreadWorkerCallback := ThreadWorkerCallback;
        if FSearchThreadCount < 1 then begin
            FMultiThread2.ThreadCount := tcAutomaticAllCoresMinus1;
        end else begin
            FMultiThread2.ThreadCount := tcManual;
            if FSearchThreadCount > 1 then begin
                FMultiThread2.ThreadCountManual := FSearchThreadCount;
            end else begin
                FMultiThread2.ThreadCountManual := 1;
            end;
        end;
        FMultiThread2.WorkMessageQueue.SetCapacity(1000);
        FMultiThread2.AutoFreeData := True;
        FMultiThread2.AutoFreeParameters := False;
        FMultiThread2.Parameters := FSearchOptions;
        FMultiThread2.Start;
    end;
end;

constructor TFOSearchThread.Create(PIDL: PItemIDList; SearchOptions: TFOSearchOptions; SearchThreadCount: Integer; Handle: HWND);
begin
    inherited Create;
    FPIDL := ILClone(PIDL);
    FSearchOptions := SearchOptions;
    FSearchOptions.InternalData.SearchThread := Self;
    FSearchThreadCount := SearchThreadCount;
    FHandle := Handle;
end;

destructor TFOSearchThread.Destroy;
begin
    FreeAndNil(FSearchOptions);
    CoTaskMemFree(FPIDL);
    inherited;
end;

procedure TFOSearchThread.Execute;
begin
    inherited;
    SearchFolderPIDL(FPIDL, FSearchOptions, FHandle, Self);
    if Assigned(FMultiThread2) then begin
        while NOT FMultiThread2.WorkFinished do begin
            Sleep(15);
            if CheckTerminated then begin
                Break;
            end;
        end;
        //FMultiThread2.Clear;
        //FMultiThread2.TerminateAllThreads;
        FreeAndNil(FMultiThread2);
    end;
    FSearchOptions.SearchFinishedCallback;
end;

procedure TFOSearchThread.ThreadWorkerCallback(Thread: TMultiThreadThread; var Parameters, Job, Results: TObject);
var
    SearchOptions: TFOSearchOptions;
    SearchScanJob: TSearchScanJob;
begin
    SearchOptions := Parameters as TFOSearchOptions;
    SearchScanJob := Job as TSearchScanJob;
    SearchCheckFileContents(SearchScanJob.FileName, SearchOptions, Thread);
end;

end.
